# Licensed under the GPLv3 - see LICENSE
"""Analysis tasks.

The tasks are imported via plugins discovered via entry point
'baseband.tasks', such as are provided by the baseband-tasks_
package.  A special rule for the entry points is that if it points
to ``__all__``, all items from that list will be imported.
Furthermore, if any name starts with '_', it is not imported.

Sample entry point entries::

    # Import just a module
    dispersion = baseband_tasks.dispersion
    # Import just one class
    Dedisperse = baseband_tasks.dispersion:Dedisperse
    # Import all entries as well as the module
    dispersion = baseband_tasks.dispersion:__all__
    # Import all entries, but not the module
    _ = baseband_tasks.dispersion:__all__

"""


def _get_entry_points():
    """Get baseband.tasks entry point.

    A list of entries where loading raised an exception are store inside
    the result under '_bad_entries'.
    """
    from entrypoints import get_group_all
    from importlib import import_module

    entries = {'_bad_entries': []}

    for entry_point in get_group_all('baseband.tasks'):
        try:
            loaded = entry_point.load()
            if entry_point.object_name == '__all__':
                module = import_module(entry_point.module_name)
                entries.update({name: getattr(module, name) for name in loaded
                                if not name.startswith('_')})
                # Possibly load module too, depending on entry point name.
                loaded = module

            if not entry_point.name.startswith('_'):
                entries[entry_point.name] = loaded
        except Exception:
            entries['_bad_entries'].append(entry_point)

    return entries


def __getattr__(attr):
    msg = f"module {__name__!r} has no attribute {attr!r}."
    if not attr.startswith('_'):
        entries = _get_entry_points()
        globals().update(entries)
        if attr in globals():
            return globals()[attr]

        if set(entries) == {'_bad_entries'}:
            msg += ('\nNo baseband.tasks entry points found. '
                    'Maybe baseband-tasks is not installed?')

    raise AttributeError(msg)


def __dir__():
    globals().update(_get_entry_points())
    return sorted(globals())



# Licensed under the GPLv3 - see LICENSE
import copy
import pickle

import pytest
import numpy as np
import astropy.units as u

from ... import guppi
from ...helpers import sequentialfile as sf
from ..base import GUPPIFileNameSequencer
from ...data import SAMPLE_PUPPI as SAMPLE_FILE


class TestGUPPI:
    def setup_class(cls):
        with open(SAMPLE_FILE, 'rb') as fh:
            cls.header = guppi.GUPPIHeader.fromfile(fh)
            cls.payload = guppi.GUPPIPayload.fromfile(fh, cls.header,
                                                      memmap=False)
        # Create a header with no overlap for stream writers.
        cls.header_w = cls.header.copy()
        cls.header_w.overlap = 0
        cls.header_w.payload_nbytes = cls.header.payload_nbytes - (
            cls.header._bpcs * cls.header.overlap // 8)

    def test_header(self, tmpdir):
        with open(SAMPLE_FILE, 'rb') as fh:
            header = guppi.GUPPIHeader.fromfile(fh)
            assert header.nbytes == 6400
            assert fh.tell() == 6400
        assert header['OBSNCHAN'] == 4
        assert header['STT_IMJD'] == 58132
        assert header['STT_SMJD'] == 51093
        assert header['STT_OFFS'] == 0
        assert header['PKTIDX'] == 0
        assert header['PKTSIZE'] == 1024
        assert header.time.isot == '2018-01-14T14:11:33.000'
        assert header.payload_nbytes == 16384
        assert header.frame_nbytes == 16384 + 6400
        assert header.overlap == 64
        assert header.samples_per_frame == 1024
        assert header.mutable is False
        with pytest.raises(TypeError):
            header['OBSNCHAN'] = 213
        assert header['OBSNCHAN'] == 4
        with pytest.raises(AttributeError):
            header.python

        with open(str(tmpdir.join('testguppi1.raw')), 'w+b') as s:
            header.tofile(s)
            assert s.tell() == header.nbytes
            s.seek(0)
            header2 = guppi.GUPPIHeader.fromfile(s)
            assert header2 == header
            assert header2.mutable is False
            assert s.tell() == header.nbytes

        # A short header raises EOFError
        with open(SAMPLE_FILE, 'rb') as fh, \
                open(str(tmpdir.join('testguppi2.raw')), 'w+b') as s:
            # Create file without "END" in first header.
            s.write(fh.read(6320))
            s.seek(0)
            # Try reading in file.
            with pytest.raises(EOFError):
                guppi.GUPPIHeader.fromfile(s)

        # A header missing "END" raises a decoding error.
        with open(SAMPLE_FILE, 'rb') as fh, \
                open(str(tmpdir.join('testguppi3.raw')), 'w+b') as s:
            # Create file without "END" in first header.
            s.write(fh.read(6320))
            fh.seek(6400)
            s.write(fh.read(10000))
            s.seek(0)
            # Try reading in file.
            with pytest.raises(UnicodeDecodeError):
                guppi.GUPPIHeader.fromfile(s)

        # Note that this is not guaranteed to preserve order!
        header3 = guppi.GUPPIHeader.fromkeys(**header)
        assert header3 == header
        assert header3.mutable is True
        # Check setting attributes.
        header3.start_time = header.start_time - 0.5 * u.day
        assert np.abs(header3.start_time
                      - (header.start_time - 0.5 * u.day)) < 1 * u.ns
        assert np.abs(header3.time - (header.time - 0.5 * u.day)) < 1 * u.ns
        header3.frame_nbytes = 13000
        assert header3.payload_nbytes == 6600
        header4 = guppi.GUPPIHeader.fromkeys(**header)
        packet_time = (header4['PKTSIZE'] * 8
                       // header4['OBSNCHAN']
                       // header4['NPOL']
                       // header4.bps) * header4['TBIN'] * u.s
        header4.offset += packet_time
        assert header4['PKTIDX'] == header['PKTIDX'] + 1
        assert np.abs(header4.time
                      - (header.time + packet_time)) < 1 * u.ns

        header5 = guppi.GUPPIHeader.fromvalues(
            start_time=header.start_time,
            sample_rate=header.sample_rate,
            samples_per_frame=header.samples_per_frame,
            overlap=header.overlap, sample_shape=header.sample_shape,
            sideband=header.sideband, bps=header.bps,
            stt_offs=header['STT_OFFS'], pktidx=header['PKTIDX'],
            pktsize=header['PKTSIZE'], obsfreq=header['OBSFREQ'],
            src_name=header['SRC_NAME'], observer=header['OBSERVER'],
            telescop=header['TELESCOP'], ra_str=header['RA_STR'],
            dec_str=header['DEC_STR'])
        # There's a lot of extraneous stuff in the headers, so only test
        # values that were passed.
        assert header5.mutable is True
        assert header5.start_time == header.start_time
        # Tests PKTSIZE, etc. all at once.
        assert header5.offset == header.offset
        assert header5.sample_rate == header.sample_rate
        assert header5.overlap == header.overlap
        assert header5.samples_per_frame == header.samples_per_frame
        assert header5.sample_shape == header.sample_shape
        assert header5.sideband == header.sideband
        assert header5.bps == header.bps
        assert header5['OBSFREQ'] == header['OBSFREQ']
        assert header5['SRC_NAME'] == header['SRC_NAME']
        assert header5['OBSERVER'] == header['OBSERVER']
        assert header5['TELESCOP'] == header['TELESCOP']
        assert header5['RA_STR'] == header['RA_STR']
        assert header5['DEC_STR'] == header['DEC_STR']

        header_tuple = (((key, header[key]) for key in header))
        header6 = guppi.GUPPIHeader(header_tuple)
        assert header6 == header

        header7 = header.copy()
        assert header7 == header
        assert header7.mutable is True
        header8 = copy.copy(header)
        assert header8 == header
        assert header8.mutable is True

        # Check that we can set the offset by either passing the offset or the
        # start time.
        offset = 9.472 * u.s
        header9 = guppi.GUPPIHeader.fromvalues(
            time=header.start_time + offset, offset=offset,
            sample_rate=header.sample_rate,
            samples_per_frame=header.samples_per_frame,
            overlap=header.overlap, sample_shape=header.sample_shape,
            sideband=header.sideband, bps=header.bps,
            pktsize=header['PKTSIZE'], obsfreq=header['OBSFREQ'],
            src_name=header['SRC_NAME'], observer=header['OBSERVER'],
            telescop=header['TELESCOP'], ra_str=header['RA_STR'],
            dec_str=header['DEC_STR'])
        header10 = guppi.GUPPIHeader.fromvalues(
            start_time=header.start_time, time=header.start_time + offset,
            sample_rate=header.sample_rate,
            samples_per_frame=header.samples_per_frame,
            overlap=header.overlap, sample_shape=header.sample_shape,
            sideband=header.sideband, bps=header.bps,
            pktsize=header['PKTSIZE'], obsfreq=header['OBSFREQ'],
            src_name=header['SRC_NAME'], observer=header['OBSERVER'],
            telescop=header['TELESCOP'], ra_str=header['RA_STR'],
            dec_str=header['DEC_STR'])
        assert np.abs(header9.offset - offset) < 1 * u.ns
        assert np.abs(header9.start_time - header.start_time) < 1 * u.ns
        assert np.abs(header10.offset - offset) < 1 * u.ns
        assert np.abs(header10.start_time - header.start_time) < 1 * u.ns

    def test_header_impossible_samples_per_frame(self):
        with pytest.raises(ValueError):
            guppi.GUPPIHeader.fromvalues(nchan=1, npol=1, bps=4,
                                         samples_per_frame=10001)

    def test_header_comment_cards(self, tmpdir):
        # Actually not obvious GUPPI itself allows it, but we might as
        # well test that we do not fail in setting comments.
        with open(SAMPLE_FILE, 'rb') as fh:
            header = guppi.GUPPIHeader.fromfile(fh)
        assert 'OBSNCHAN' not in header.comments
        header1 = header.copy()
        header1['OBSNCHAN'] = header['OBSNCHAN'], 'number of channels'
        assert header1.comments['OBSNCHAN'] == 'number of channels'
        name = str(tmpdir.join('guppi_header.test'))
        with open(name, 'wb') as fw:
            header1.tofile(fw)

        with open(name, 'rb') as fr:
            header2 = guppi.GUPPIHeader.fromfile(fr)

        assert header2 == header
        assert header2.comments['OBSNCHAN'] == 'number of channels'

    def test_payload(self, tmpdir):
        payload = self.payload
        assert payload.nbytes == 16384
        assert payload.shape == (1024, 2, 4)
        # Check sample shape validity.
        assert payload.sample_shape == (2, 4)
        assert payload.sample_shape.npol == 2
        assert payload.sample_shape.nchan == 4
        assert payload.size == 8192
        assert payload.ndim == 3
        assert payload.dtype == np.complex64
        assert np.all(payload[:3] == np.array(
            [[[-7.+12.j, -32.-10.j, -17.+25.j, 16.-5.j],
              [14.+21.j, -5.-7.j, 19.-8.j, 7.+7.j]],
             [[5.-3.j, -15.-14.j, -8.+14.j, -6.-18.j],
              [21.-1.j, 22.+6.j, -30.-13.j, 12.+23.j]],
             [[11.+2.j, 9.-13.j, 9.-15.j, -21.-6.j],
              [10.-12.j, -3.-10.j, -12.-8.j, 4.-27.j]]], dtype=np.complex64))
        assert np.all(payload[337:340] == np.array(
            [[[2.-25.j, 31.+2.j, -10.+1.j, -29.+14.j],
              [24.+6.j, -23.-16.j, -22.-20.j, -11.-6.j]],
             [[11.+10.j, -2.-1.j, -6.+9.j, 19.+16.j],
              [10.-25.j, -33.-5.j, 14.+0.j, 3.-3.j]],
             [[22.-7.j, 5.+11.j, -21.+4.j, 2.+0.j],
              [-4.-12.j, 1.+1.j, 13.+6.j, -31.-4.j]]], dtype=np.complex64))
        data = payload.data
        assert np.all(payload[:] == data)

        with open(str(tmpdir.join('testguppi.raw')), 'w+b') as s:
            payload.tofile(s)
            s.seek(0)
            payload2 = guppi.GUPPIPayload.fromfile(s, payload_nbytes=16384,
                                                   sample_shape=(2, 4), bps=8,
                                                   complex_data=True)
            assert s.tell() == 16384
            assert payload2 == payload
            with pytest.raises(EOFError):
                # Too few bytes.
                s.seek(100)
                guppi.GUPPIPayload.fromfile(s, self.header, memmap=False)

        payload3 = guppi.GUPPIPayload.fromdata(payload.data, bps=8)
        assert payload3 == payload

        with open(SAMPLE_FILE, 'rb') as fh:
            fh.seek(self.header.nbytes)
            payload4 = guppi.GUPPIPayload.fromfile(fh, self.header,
                                                   memmap=True)
        assert isinstance(payload4.words, np.memmap)
        assert not isinstance(payload.words, np.memmap)
        assert payload == payload4

        # Check selective writing.
        payload5 = guppi.GUPPIPayload.fromdata(payload.data, bps=8)
        payload5[547:563, 0, :3] = (-1. + 3.j)
        assert np.all(payload5[547:563, 0, :3] == (-1. + 3.j))
        # Check nothing else has changed.
        assert np.all(payload5[:547] == payload[:547])
        assert np.all(payload5[563:] == payload[563:])
        assert np.all(payload5[547:563, 1] == payload[547:563, 1])
        assert np.all(payload5[547:563, 0, 3] == payload[547:563, 0, 3])
        some_data = np.array([5.-4.j, -2.+8.j], dtype=np.complex64)
        payload5[11:13, 1, 2] = some_data
        assert np.all(payload5[11:13, 1, 2] == some_data)
        # We can't yet read negative steps.
        with pytest.raises(AssertionError) as excinfo:
            payload5[27:13:-1, 1, 2]
        assert "cannot deal with negative steps" in str(excinfo.value)

        # Check that we can create and decode (nsample, nchan, npol) payloads.
        payload_tfirst = guppi.GUPPIPayload.fromdata(self.payload.data, bps=8,
                                                     channels_first=False)
        # First check channels_first makes a difference to payload words.
        assert ~np.all(payload_tfirst.words == payload.words)
        # Now check the data is the same.
        assert np.all(payload_tfirst.data == payload.data)
        # Check selective decode.
        item = (slice(547, 829, 2), slice(None), np.array([2, 1]))
        assert np.all(payload_tfirst[item] == payload[item])

        with pytest.raises(ValueError, match='cannot encode'):
            guppi.GUPPIPayload.fromdata(payload.data, bps=4)

    def test_file_reader(self):
        with guppi.open(SAMPLE_FILE, 'rb') as fh:
            header = fh.read_header()
            assert header == self.header
            current_pos = fh.tell()
            frame_rate = fh.get_frame_rate()
            assert frame_rate == (header.sample_rate
                                  / (header.samples_per_frame
                                     - header.overlap))
            assert fh.tell() == current_pos
            # Frame is done below, as is writing in binary.

    def test_file_info(self):
        with guppi.open(SAMPLE_FILE, 'rb') as fh:
            info = fh.info
            assert info.format == 'guppi'
            header = fh.read_header()
            assert info.bps == header.bps
            assert info.complex_data == header.complex_data
            assert info.sample_shape == header.sample_shape
            assert info.start_time == header.start_time
            assert info.samples_per_frame == header.samples_per_frame
            assert info.overlap == header.overlap
            # Note: sample_rate used to be wrong, since calculated from
            # from frame rate and samples_per_frame w/o corr. for overlap.
            assert info.sample_rate == header.sample_rate
            assert info.frame_rate == header.sample_rate / (
                header.samples_per_frame - header.overlap)

    def test_file_info_unsupported_format(self, tmpdir):
        filename = str(tmpdir.join('file.uppi'))
        with guppi.open(SAMPLE_FILE, 'rb') as fh:
            f = fh.read_frame()
            f.header = f.header.copy()
            f['PKTFMT'] = 'unknown'
            with guppi.open(filename, 'wb') as fw:
                fw.write_frame(f)

        with guppi.open(filename, 'rb') as fr:
            info = fr.info

        assert info.pktfmt == 'unknown'
        assert 'pktfmt' in info.warnings
        assert 'Unknown pktfmt' in info.warnings['pktfmt']

    def test_frame(self, tmpdir):
        with guppi.open(SAMPLE_FILE, 'rb') as fh:
            frame = fh.read_frame(memmap=False)
            assert fh.tell() == frame.nbytes
        header, payload = frame.header, frame.payload
        assert header == self.header
        assert payload == self.payload
        assert frame == guppi.GUPPIFrame(header, payload)
        assert frame.sample_shape == payload.sample_shape
        assert frame.shape == (len(frame),) + frame.sample_shape
        assert frame.size == len(frame) * np.prod(frame.sample_shape)
        assert frame.ndim == payload.ndim
        assert np.all(frame[337:340] == np.array(
            [[[2.-25.j, 31.+2.j, -10.+1.j, -29.+14.j],
              [24.+6.j, -23.-16.j, -22.-20.j, -11.-6.j]],
             [[11.+10.j, -2.-1.j, -6.+9.j, 19.+16.j],
              [10.-25.j, -33.-5.j, 14.+0.j, 3.-3.j]],
             [[22.-7.j, 5.+11.j, -21.+4.j, 2.+0.j],
              [-4.-12.j, 1.+1.j, 13.+6.j, -31.-4.j]]], dtype=np.complex64))

        with open(str(tmpdir.join('testguppi.raw')), 'w+b') as s:
            frame.tofile(s)
            s.seek(0)
            frame2 = guppi.GUPPIFrame.fromfile(s, memmap=False)
        assert frame2 == frame
        frame3 = guppi.GUPPIFrame.fromdata(payload.data, header)
        assert frame3 == frame
        frame4 = guppi.GUPPIFrame.fromdata(payload.data, **header)
        assert frame4 == frame
        header5 = header.copy()
        frame5 = guppi.GUPPIFrame(header5, payload, valid=False)
        assert frame5.valid is False
        assert np.all(frame5.data == 0.)
        invalid_samples = frame5[-1000:]
        assert np.all(invalid_samples == 0.)
        assert invalid_samples.shape == (1000, 2, 4)
        invalid_samples = frame5[8192:]
        assert invalid_samples.shape == (0, 2, 4)
        frame5.valid = True
        assert frame5 == frame

    def test_frame_memmap(self, tmpdir):
        # Get frame the regular way.
        with guppi.open(SAMPLE_FILE, 'rb') as fr:
            frame = fr.read_frame(memmap=False)
        assert not isinstance(frame.payload.words, np.memmap)
        # Check that if we map it instead, we get the same result.
        with guppi.open(SAMPLE_FILE, 'rb') as fh:
            frame2 = fh.read_frame(memmap=True)
        assert frame2 == frame
        assert isinstance(frame2.payload.words, np.memmap)
        # Bit superfluous perhaps, but check decoding as well.
        assert np.all(frame2[337:340] == np.array(
            [[[2.-25.j, 31.+2.j, -10.+1.j, -29.+14.j],
              [24.+6.j, -23.-16.j, -22.-20.j, -11.-6.j]],
             [[11.+10.j, -2.-1.j, -6.+9.j, 19.+16.j],
              [10.-25.j, -33.-5.j, 14.+0.j, 3.-3.j]],
             [[22.-7.j, 5.+11.j, -21.+4.j, 2.+0.j],
              [-4.-12.j, 1.+1.j, 13.+6.j, -31.-4.j]]], dtype=np.complex64))
        assert np.all(frame2.data == frame.data)

        # Now check writing.  First, without memmap, just ensuring writing
        # to file works as well as to BytesIO done above.
        filename = str(tmpdir.join('testguppi.raw'))
        with guppi.open(filename, 'wb') as fw:
            fw.write_frame(frame)

        with guppi.open(filename, 'rb') as fw:
            frame3 = fw.read_frame()

        assert frame3 == frame

        # Now memmap file to be written to.
        # Use a new file since the mmap used in frame3 may still have a handle
        # to the previous one, which causes (sensible) failures on windows.
        # See https://bugs.python.org/issue40720
        filename2 = str(tmpdir.join('testguppi2.raw'))
        with guppi.open(filename2, 'wb') as fw:
            frame4 = fw.memmap_frame(frame.header)
        # Initially no data set, so frames should not match yet.
        assert frame4 != frame
        # So, if we read this file, it also should not match.
        with guppi.open(filename2, 'rb') as fw:
            frame5 = fw.read_frame()
        assert frame5 != frame

        # Fill in some data.  This should only update some words.
        frame4[:20] = frame[:20]
        assert np.all(frame4[:20] == frame[:20])
        assert frame4 != frame
        # Update the rest, so it becomes the same.
        frame4[20:] = frame[20:]
        assert frame4 == frame
        # Delete to flush to disk just to be sure, then read and check it.
        del frame4
        with guppi.open(filename2, 'rb') as fn:
            frame6 = fn.read_frame()

        assert frame6 == frame
        # Some further tests for completeness;
        # initiate frame using data and header keywords.
        filename3 = str(tmpdir.join('testguppi3.raw'))
        with guppi.open(filename3, 'wb') as fw:
            fw.write_frame(self.payload.data, **self.header)
        with guppi.open(filename3, 'rb') as fh:
            frame7 = fh.read_frame()
        assert frame7 == frame
        # memmap frame using header keywords.
        filename4 = str(tmpdir.join('testguppi4.raw'))
        with guppi.open(filename4, 'wb') as fw:
            frame8 = fw.memmap_frame(**self.header)
            frame8[:] = self.payload.data
        assert frame8 == frame
        del frame8
        with guppi.open(filename4, 'rb') as fh:
            frame9 = fh.read_frame()
        assert frame9 == frame

    def test_filestreamer(self, tmpdir):
        start_time = self.header.time
        nsample = 4*1024 - 3*64
        with guppi.open(SAMPLE_FILE, 'rs') as fh:
            assert fh.header0 == self.header
            assert fh.sample_shape == (2, 4)
            assert fh.shape == (nsample,) + fh.sample_shape
            assert fh.size == np.prod(fh.shape)
            assert fh.ndim == len(fh.shape)
            assert fh.start_time == start_time
            assert fh.sample_rate == 250 * u.Hz
            record = fh.read()
            fh.seek(0)
            record1 = fh.read(12)
            assert fh.tell() == 12
            fh.seek(1523)
            record2 = np.zeros((2, 2, 4), dtype=np.complex64)
            record2 = fh.read(out=record2)
            # Check that stream properly skips overlap.
            assert np.all(record2 == fh._frame[563:565])
            assert fh.tell() == 1525
            assert fh.time == fh.tell(unit='time')
            assert (np.abs(fh.time - (start_time + 1525 / (250 * u.Hz)))
                    < 1. * u.ns)
            fh.seek(fh.start_time + 100 / (250 * u.Hz))
            assert fh.tell() == 100
            assert (np.abs(fh.stop_time
                           - (start_time + nsample / (250 * u.Hz))) < 1.*u.ns)
            fh.seek(1, 'end')
            with pytest.raises(EOFError):
                fh.read()

        assert record1.shape == (12, 2, 4)
        assert np.all(record1[:3] == np.array(
            [[[-7.+12.j, -32.-10.j, -17.+25.j, 16.-5.j],
              [14.+21.j, -5.-7.j, 19.-8.j, 7.+7.j]],
             [[5.-3.j, -15.-14.j, -8.+14.j, -6.-18.j],
              [21.-1.j, 22.+6.j, -30.-13.j, 12.+23.j]],
             [[11.+2.j, 9.-13.j, 9.-15.j, -21.-6.j],
              [10.-12.j, -3.-10.j, -12.-8.j, 4.-27.j]]], dtype=np.complex64))
        assert record1.dtype == np.complex64
        assert np.all(record1 == self.payload[:12].squeeze())
        assert record2.shape == (2, 2, 4)

        filename = str(tmpdir.join('testguppi.raw'))
        spf = self.header.samples_per_frame - self.header.overlap
        # Need to remove overlap.
        with guppi.open(filename, 'ws', header0=self.header_w,
                        squeeze=False) as fw:
            assert fw.sample_rate == 250 * u.Hz
            fw.write(self.payload[:spf])
            assert fw.start_time == start_time
            assert (np.abs(fw.time - (start_time + spf / (250 * u.Hz)))
                    < 1. * u.ns)

        with guppi.open(filename, 'rs') as fh:
            data = fh.read()
            assert fh.start_time == start_time
            assert (np.abs(fh.time - (start_time + spf / (250 * u.Hz)))
                    < 1. * u.ns)
            assert fh.stop_time == fh.time
            assert fh.sample_rate == 250 * u.Hz
        assert np.all(data == self.payload[:spf].squeeze())

        # Try single polarisation (two channels to keep it complex), and check
        # initialisation by header keywords.  This also tests writing
        # with squeeze=True.
        h = self.header
        filename2 = str(tmpdir.join('testguppi2.raw'))
        with guppi.open(filename2, 'ws', time=h.time, bps=h.bps,
                        sample_rate=h.sample_rate,
                        pktsize=h['PKTSIZE'], overlap=0,
                        payload_nbytes=self.header_w.payload_nbytes // 4,
                        nchan=2, npol=1) as fw:
            fw.write(self.payload[:spf, 0, :2])
            assert np.abs(fw.start_time - start_time) < 1.*u.ns
            assert (np.abs(fw.time - (start_time + spf / (250 * u.Hz)))
                    < 1. * u.ns)

        with guppi.open(filename2, 'rs') as fh:
            data_onepol = fh.read()
            assert np.abs(fh.start_time - start_time) < 1.*u.ns
            assert np.abs(fh.stop_time
                          - (start_time + spf / (250 * u.Hz))) < 1.*u.ns
        assert np.all(data_onepol == self.payload[:spf, 0, :2])

        # Try reading a single polarization.
        with guppi.open(SAMPLE_FILE, 'rs', subset=0) as fh:
            assert fh.sample_shape == (4,)
            assert fh.subset == (0,)
            record3 = fh.read(12)
            assert np.all(record3 == record[:12, 0, :4])

        # Read right polarization, but read channels in reverse order.
        with guppi.open(SAMPLE_FILE, 'rs', subset=(1, [1, 0])) as fh:
            assert fh.sample_shape == (2,)
            data_sub = fh.read()
            assert np.all(data_sub[:, 1] == record[:, 1, 0])

        # Test that squeeze attribute works on read (including in-place read).
        with guppi.open(filename2, 'rs', squeeze=False) as fh:
            assert fh.sample_shape == (1, 2)
            assert fh.sample_shape.npol == 1
            assert fh.sample_shape.nchan == 2
            assert fh.read(1).shape == (1, 1, 2)
            assert fh.read(10).shape == (10, 1, 2)
            fh.seek(0)
            out = np.zeros((12, 1, 2), dtype=np.complex64)
            fh.read(out=out)
            assert fh.tell() == 12
            assert np.all(out.squeeze() == self.payload[:12, 0, :2])

        # Test that squeeze=False works on write (we checked above it works
        # with squeeze=True).
        filename3 = str(tmpdir.join('testguppi3.raw'))
        with guppi.open(filename3, 'ws', time=h.time, bps=h.bps,
                        sample_rate=h.sample_rate,
                        pktsize=h['PKTSIZE'], overlap=0,
                        payload_nbytes=self.header_w.payload_nbytes // 4,
                        nchan=2, npol=1, squeeze=False) as fw:
            fw.write(self.payload[:spf, 0:1, :2])
            assert np.abs(fw.start_time - start_time) < 1.*u.ns
            assert (np.abs(fw.time - (start_time + spf / (250 * u.Hz)))
                    < 1. * u.ns)

        with guppi.open(filename3, 'rs', squeeze=False) as fh:
            data_onepol = fh.read()
        assert np.all(data_onepol.squeeze() == self.payload[:spf, 0, :2])

    def test_stream_overlap(self):
        # See gh-264
        with guppi.open(SAMPLE_FILE, 'rs') as fh:
            overlap = fh.header0.overlap
            # Check "normal" end.
            fh.seek(4*fh.samples_per_frame)
            data = fh.read()
            assert len(data) == overlap
            fh.seek(-1, 2)
            assert fh.tell() == 4*fh.samples_per_frame + overlap - 1
            data = fh.read()
            assert len(data) == 1

    def test_incomplete_stream(self, tmpdir):
        """Test that writing an incomplete stream is possible, and that frame
        set is valid but invalid samples use the fill value.
        """
        filename = str(tmpdir.join('testguppi.raw'))
        with pytest.warns(UserWarning, match='partial buffer'):
            with guppi.open(filename, 'ws', header0=self.header_w,
                            squeeze=False) as fw:
                fw.write(self.payload[:10])

        with guppi.open(filename, 'rs', squeeze=False) as fwr:
            data = fwr.read()
            assert np.all(data[:10] == self.payload[:10])
            assert np.all(data[10:] == fwr.fill_value)

    def test_pickle(self):
        # Only simple tests here; more complete ones in vdif.
        with guppi.open(SAMPLE_FILE, 'rs') as fh:
            fh.seek(6)
            pickled = pickle.dumps(fh)
            fh.read(3)
            with pickle.loads(pickled) as fh2:
                assert fh2.tell() == 6
                fh2.read(10)

            assert fh.tell() == 9

        with pickle.loads(pickled) as fh3:
            assert fh3.tell() == 6
            fh3.read(1)

        closed = pickle.dumps(fh)
        with pickle.loads(closed) as fh4:
            assert fh4.closed
            with pytest.raises(ValueError):
                fh4.read(1)

    def test_multiple_files_stream(self, tmpdir):
        start_time = self.header_w.time
        with guppi.open(SAMPLE_FILE, 'rs') as fh:
            data = fh.read(3840)  # omit overlap for test
        filenames = (str(tmpdir.join('guppi_1.raw')),
                     str(tmpdir.join('guppi_2.raw')))
        with guppi.open(filenames, 'ws', header0=self.header_w,
                        frames_per_file=2) as fw:
            start_time = fw.start_time
            fw.write(data[:1000])
            time1000 = fw.time
            fw.write(data[1000:])
            stop_time = fw.time
        assert start_time == self.header_w.time
        assert np.abs(time1000
                      - (start_time + 1000 / (250 * u.Hz))) < 1.*u.ns
        assert np.abs(stop_time
                      - (start_time + 3840 / (250 * u.Hz))) < 1.*u.ns

        with guppi.open(filenames[1], 'rs') as fr:
            assert (np.abs(fr.time - (start_time + 1920 / (250 * u.Hz)))
                    < 1. * u.ns)
            data1 = fr.read()
        assert np.all(data1 == data[1920:])

        with guppi.open(filenames, 'rs') as fr:
            assert fr.start_time == start_time
            assert fr.time == start_time
            assert np.abs(fr.stop_time
                          - (start_time + 3840 / (250 * u.Hz))) < 1. * u.ns
            data2 = fr.read()
            assert np.abs(fr.time - fr.stop_time) < 1. * u.ns
        assert np.all(data2 == data)

        # Pass sequentialfile objects to reader.
        filenames = (str(tmpdir.join('guppi2_1.raw')),
                     str(tmpdir.join('guppi2_2.raw')))
        with sf.open(filenames, 'w+b',
                     file_size=2*self.header_w.frame_nbytes) as fraw, \
                guppi.open(fraw, 'ws', header0=self.header_w) as fw:
            fw.write(data)

        with sf.open(filenames, 'rb') as fraw, \
                guppi.open(fraw, 'rs') as fr:
            data3 = fr.read()
            # Test pickling in the process
            pickled = pickle.dumps(fr)
        assert np.all(data3 == data)

        with pickle.loads(pickled) as fr2:
            assert fr2.tell() == fr2.shape[0]
            fr2.seek(-10, 2)
            datap = fr2.read()
        assert np.all(datap.squeeze() == data[-10:])

        # Check that we can't pass a filename sequence in 'wb' mode.
        with pytest.raises(ValueError):
            guppi.open(filenames, 'wb')

    def test_partial_last_frame(self, tmpdir):
        """Test reading a file with an incomplete last frame."""
        # Read in sample file as a byte stream.
        with guppi.open(SAMPLE_FILE, 'rb') as fh:
            puppi_raw = fh.read()

        # Try reading a file with an incomplete payload.
        with guppi.open(str(tmpdir.join('puppi_partframe.raw')), 'wb') as fw:
            fw.write(puppi_raw[:len(puppi_raw) - 6091])
        nsample = 3*1024 - 2*64  # 3 frames minus 2 overlaps
        with guppi.open(str(tmpdir.join('puppi_partframe.raw')), 'rs') as fn:
            assert fn.shape == (nsample, 2, 4)
            assert np.abs(fn.stop_time
                          - fn.start_time - nsample / (250 * u.Hz)) < 1. * u.ns

        # Try reading a file with an incomplete header.
        with guppi.open(str(tmpdir.join('puppi_partframe.raw')), 'wb') as fw:
            fw.write(puppi_raw[:len(puppi_raw) - 17605])
        with guppi.open(str(tmpdir.join('puppi_partframe.raw')), 'rs') as fn:
            assert fn.shape == (nsample, 2, 4)
            assert np.abs(fn.stop_time
                          - fn.start_time - nsample / (250 * u.Hz)) < 1. * u.ns

    def test_chan_ordered_stream(self, tmpdir):
        """Test encoding and decoding frames and streams that use
        (nsample, nchan, npol).
        """
        filename = str(tmpdir.join('testguppi.raw'))

        with guppi.open(SAMPLE_FILE) as fh:
            data = fh.read(3840)  # omit overlap for test

        header = self.header.copy()
        header.channels_first = False
        header['OVERLAP'] = 0
        header.samples_per_frame = 960
        with guppi.open(filename, 'ws', header0=header) as fw:
            fw.write(data)
        with guppi.open(filename) as fn:
            fn.seek(1231)
            new_data = fn.read(47)
            assert np.all(new_data == data[1231:1231 + 47])

    def test_template_stream(self, tmpdir):
        start_time = self.header_w.time
        with guppi.open(SAMPLE_FILE, 'rs') as fh:
            data = fh.read(3840)  # omit overlap for test

        # Simple template with file number counter.
        template = str(tmpdir.join('guppi_{file_nr:02d}.raw'))
        with guppi.open(template, 'ws', frames_per_file=1,
                        **self.header_w) as fw:
            fw.write(data)

        with guppi.open(template, 'rs') as fr:
            assert len(fr.fh_raw.files) == 4
            assert fr.fh_raw.files[-1] == str(tmpdir.join('guppi_03.raw'))
            assert np.abs(fr.stop_time
                          - (start_time + 3840 / (250 * u.Hz))) < 1.*u.ns
            data2 = fr.read()
        assert np.all(data2 == data)

        # More complex template that requires keywords.
        template = str(tmpdir.join('puppi_{stt_imjd}.{file_nr:04d}.raw'))
        with guppi.open(template, 'ws', frames_per_file=1,
                        header0=self.header_w) as fw:
            fw.write(data[:1920])
            assert fw.start_time == start_time
            assert (np.abs(fw.time - (start_time + 1920 / (250 * u.Hz)))
                    < 1. * u.ns)
            fw.write(data[1920:])
            assert (np.abs(fw.time - (start_time + 3840 / (250 * u.Hz)))
                    < 1. * u.ns)

        # We cannot just open using the same template, since STT_IMJD is
        # not available.
        with pytest.raises(KeyError):
            guppi.open(template, 'rs')

        kwargs = dict(STT_IMJD=self.header_w['STT_IMJD'])
        with guppi.open(template, 'rs', **kwargs) as fr:
            data3 = fr.read()
        assert np.all(data3 == data)

        # Try passing stream reader kwargs.
        with guppi.open(template, 'rs', subset=(0, [2, 3]), squeeze=False,
                        **kwargs) as fr:
            data4 = fr.read()
        assert np.all(data4.squeeze() == data[:, 0, 2:])

        # Just to check internal checks are OK.
        filename = template.format(stt_imjd=self.header_w['STT_IMJD'],
                                   file_nr=0)
        with pytest.raises(ValueError):
            guppi.open(filename, 's')
        with pytest.raises(TypeError):
            # Extraneous argument.
            guppi.open(filename, 'rs', files=(filename,))

    def test_stream_info(self):
        with guppi.open(SAMPLE_FILE, 'rs') as fh:
            info = fh.info
            assert info.format == 'guppi'
            assert info.shape == fh.shape
            assert info.sample_rate == fh.sample_rate
            assert info.start_time == fh.start_time
            assert info.stop_time == fh.stop_time
            assert info.file_info is fh.fh_raw.info


class TestGUPPIFileNameSequencer:
    def setup(self):
        with open(SAMPLE_FILE, 'rb') as fh:
            self.header = guppi.GUPPIHeader.fromfile(fh)

    def test_header_extraction(self):
        # Follow Nikhil's J1810 Arecibo observation file naming scheme:
        # puppi_58132_J1810+1744_2176.0000.raw, etc.
        template = 'puppi_{stt_imjd}_{src_name}_{scannum}.{file_nr:04d}.raw'
        fns = GUPPIFileNameSequencer(template, self.header)
        assert fns[0] == 'puppi_58132_J1810+1744_2176.0000.raw'
        assert fns[29] == 'puppi_58132_J1810+1744_2176.0029.raw'



# Licensed under the GPLv3 - see LICENSE
"""Mark5B VLBI data reader.

Code inspired by Walter Brisken's mark5access.  See
https://github.com/demorest/mark5access.

Also, for the Mark5B design, see
https://www.haystack.mit.edu/tech/vlbi/mark5/mark5_memos/019.pdf
"""
from .base import open, info  # noqa
from .header import Mark5BHeader  # noqa
from .payload import Mark5BPayload  # noqa
from .frame import Mark5BFrame  # noqa



# Licensed under the GPLv3 - see LICENSE
import pytest
import numpy as np
from astropy.time import Time
import astropy.units as u

from .. import open as baseband_open, file_info
from ..helpers import sequentialfile as sf
from ..data import (SAMPLE_MARK4 as SAMPLE_M4, SAMPLE_MARK5B as SAMPLE_M5B,
                    SAMPLE_VDIF, SAMPLE_DADA)


@pytest.mark.parametrize(
    ('sample', 'fmt'),
    ((SAMPLE_M4, 'mark4'),
     (SAMPLE_M5B, 'mark5b'),
     (SAMPLE_VDIF, 'vdif')))
def test_open(sample, fmt):
    # Use extra arguments that allow all formats to succeed.
    extra_args = {'nchan': 8,
                  'ref_time': Time('2014-01-01'),
                  'sample_rate': 32.*u.MHz}
    info = file_info(sample, fmt, **extra_args)
    with baseband_open(sample, 'rs', **extra_args) as fh:
        assert fh.start_time == info.start_time


@pytest.mark.parametrize('sample', (SAMPLE_M4, SAMPLE_M5B))
def test_open_missing_args(sample):
    with pytest.raises(TypeError) as exc:
        baseband_open(sample, 'rs')
    assert "missing required arguments" in str(exc.value)


def test_open_wrong_args():
    # Use extra arguments that allow formats to succeed up to the stream
    # reader, but then fail on an incorrect sample rate.
    mark4_args = {'nchan': 8,
                  'ref_time': Time('2014-01-01')}
    with pytest.raises(ValueError, match='inconsistent'):  # wrong sample_rate
        baseband_open(SAMPLE_M4, 'rs', sample_rate=31*u.MHz, **mark4_args)

    with pytest.raises(TypeError, match='unexpected'):  # extraneous argument
        baseband_open(SAMPLE_M4, 'rs', life=42, **mark4_args)

    with pytest.raises(ValueError, match='inconsistent'):  # wrong decade
        baseband_open(SAMPLE_VDIF, 'rs', decade=2000)

    with pytest.raises(ValueError, match='inconsistent'):  # wrong kday
        baseband_open(SAMPLE_VDIF, 'rs', kday=55000)

    with pytest.raises(ValueError, match='inconsistent'):  # ref_time off.
        baseband_open(SAMPLE_VDIF, 'rs', ref_time=Time('J2000'))

    with pytest.raises(ValueError, match='inconsistent'):  # nchan wrong.
        baseband_open(SAMPLE_DADA, 'rs', nchan=8)

    with pytest.raises(TypeError):  # decade not int
        baseband_open(SAMPLE_M4, 'rs', decade='2010')

    with pytest.raises(TypeError):  # kday not int
        baseband_open(SAMPLE_M5B, 'rs', kday='unknown', nchan=8, bps=2)


def test_open_sequence(tmpdir):
    """Test opening file sequences with `baseband.open`."""
    # Open DADA file sequence by passing a list.
    with baseband_open(SAMPLE_DADA) as fh:
        data1 = fh.read()
        header1 = fh.header0.copy()
    header1.payload_nbytes = header1.payload_nbytes // 2

    files = [str(tmpdir.join('f.{0:d}.dada'.format(x))) for x in range(2)]
    with baseband_open(files, 'ws', format='dada', header0=header1) as fw:
        fw.write(data1)

    with baseband_open(files) as fn:
        assert fn.info.format == 'dada'
        assert len(fn.fh_raw.files) == 2
        assert fn.header0 == header1
        assert np.all(fn.read() == data1)

    # Open VDIF file sequence by passing a FileNameSequencer.
    files = sf.FileNameSequencer(str(tmpdir.join('f{file_nr:03d}.vdif')))
    with baseband_open(SAMPLE_VDIF) as fh:
        data2 = fh.read()
        header2 = fh.header0.copy()

    # Sample file has 2 framesets of 8 frames each.
    with baseband_open(files, 'ws', format='vdif', nthread=8,
                       file_size=8*header2.frame_nbytes, **header2) as fw:
        fw.write(data2)

    # Can't check header0 because frameset might be out of order.
    with baseband_open(files) as fn:
        assert fn.info.format == 'vdif'
        assert len(fn.fh_raw.files) == 2
        assert np.all(data2 == fn.read())


def test_open_write_checks():
    # Cannot have multiple formats for writing.
    with pytest.raises(ValueError, match='cannot specify multiple'):
        baseband_open('a.a', 'wb', fmt=('dada', 'mark4'))


def test_unsupported_file(tmpdir):
    name = str(tmpdir.join('test.unsupported'))
    with open(name, 'wb') as fw:
        fw.write(b'abcdefghijklmnopqrstuvwxyz')

    with pytest.raises(ValueError, match='could not be auto-determined'):
        with pytest.warns(UserWarning, match='not.*formatted'):
            baseband_open(name)


def test_format_with_no_info(monkeypatch):
    monkeypatch.delattr('baseband.vdif.info')
    info = file_info(SAMPLE_VDIF)
    assert info
    assert info.format == 'vdif'
    assert not hasattr(info, 'used_kwargs')

    with baseband_open(SAMPLE_VDIF, format=('vdif', 'mark5b')) as fh:
        assert fh.info.format == 'vdif'

    with pytest.raises(ValueError, match='could not be auto-determined'):
        with pytest.warns(UserWarning, match='not.*formatted'):
            baseband_open(SAMPLE_M4, format=('vdif', 'mark5b'))



# Licensed under the GPLv3 - see LICENSE
"""The Mark4FileReaderInfo property.

Includes information about what is needed to calcuate times,
number of tracks and offset of first header.
"""

from ..base.file_info import FileReaderInfo, info_item


__all__ = ['Mark4FileReaderInfo']


class Mark4FileReaderInfo(FileReaderInfo):
    """Standardized information on Mark 4 file readers.

    The ``info`` descriptor has a number of standard attributes, which are
    determined from arguments passed in opening the file, from the first header
    (``info.header0``) and from possibly scanning the file to determine the
    duration of frames.  This class has two additional attributes specific to
    Mark 4 files (``ntrack`` and ``offset0``, see below).

    Examples
    --------
    The most common use is simply to print information::

        >>> from baseband.data import SAMPLE_MARK4
        >>> from baseband import mark4
        >>> fh = mark4.open(SAMPLE_MARK4, 'rb')
        >>> fh.info
        Mark4File information:
        format = mark4
        number_of_frames = 2
        frame_rate = 400.0 Hz
        sample_rate = 32.0 MHz
        samples_per_frame = 80000
        sample_shape = (8,)
        bps = 2
        complex_data = False
        readable = True
        ntrack = 64
        offset0 = 2696
        <BLANKLINE>
        missing:  decade, ref_time: needed to infer full times.
        <BLANKLINE>
        checks:  decodable: True
        >>> fh.close()

        >>> fh = mark4.open(SAMPLE_MARK4, 'rb', decade=2010)
        >>> fh.info
        Mark4File information:
        format = mark4
        number_of_frames = 2
        frame_rate = 400.0 Hz
        sample_rate = 32.0 MHz
        samples_per_frame = 80000
        sample_shape = (8,)
        bps = 2
        complex_data = False
        start_time = 2014-06-16T07:38:12.475000000
        readable = True
        ntrack = 64
        offset0 = 2696
        <BLANKLINE>
        checks:  decodable: True
        >>> fh.close()
    """
    attr_names = (FileReaderInfo.attr_names[:-4]
                  + ('ntrack', 'offset0')
                  + FileReaderInfo.attr_names[-4:])
    """Attributes that the container provides."""

    ntrack = info_item(needs='_parent', doc=(
        'Number of "tape tracks" simulated in the disk file.'))
    decade = info_item(needs='_parent', doc=(
        'Decade in which the observations were taken'))
    ref_time = info_item(needs='_parent', doc=(
        'Reference time within 4 years of the observation time'))

    @info_item
    def time_info(self):
        """Additional time info needed to get the start time."""
        time_info = (self.decade, self.ref_time)
        if time_info == (None, None):
            self.missing['decade'] = self.missing['ref_time'] = (
                "needed to infer full times.")
            return None

        return time_info

    @info_item
    def offset0(self):
        """Offset in bytes to the location of the first header."""
        with self._parent.temporary_offset(0) as fh:
            return fh.locate_frames()[0]

    @info_item(needs='offset0')
    def header0(self):
        with self._parent.temporary_offset(self.offset0) as fh:
            return fh.read_header()

    @info_item(needs='header0')
    def frame0(self):
        with self._parent.temporary_offset(self.offset0) as fh:
            return fh.read_frame()

    @info_item(needs='header0')
    def number_of_frames(self):
        """Total number of frames."""
        with self._parent.temporary_offset(
                -self.header0.frame_nbytes, 2) as fh:
            fh.find_header(self.header0, forward=False)
            number_of_frames = ((fh.tell() - self.offset0)
                                / self.header0.frame_nbytes) + 1

        if number_of_frames % 1 == 0:
            return int(number_of_frames)
        else:
            self.warnings['number_of_frames'] = (
                'file contains non-integer number ({}) of frames'
                .format(number_of_frames))
            return None

    # Override just to replace what it "needs".
    @info_item(needs='offset0')
    def format(self):
        return 'mark4'

    @info_item(needs=('header0', 'time_info'))
    def start_time(self):
        """Time of the first sample."""
        return super().start_time



# Licensed under the GPLv3 - see LICENSE
"""Test things that are not already tested with the many formats."""
import pytest

from ... import vdif, data
from ..file_info import FileReaderInfo, StreamReaderInfo


def test_str_repr():
    # Check names are interpreted correctly for all types of info_item
    # Directly assigned
    assert str(FileReaderInfo.errors).startswith('errors: ')
    # From parent
    assert str(StreamReaderInfo.start_time).startswith('start_time: ')
    # From header0
    assert str(FileReaderInfo.bps).startswith('bps: ')
    # From function
    assert str(FileReaderInfo.header0).startswith('header0: ')
    # From function with needs (i.e., via __call__)
    assert str(FileReaderInfo.readable).startswith('readable: ')
    assert repr(FileReaderInfo.errors).startswith('<info_item errors')
    assert repr(FileReaderInfo()).startswith('FileReaderInfo (unbound)')
    assert repr(StreamReaderInfo()).startswith('StreamReaderInfo (unbound)')
    assert repr(vdif.base.VDIFFileReader.info).startswith(
        'VDIFFileReaderInfo (unbound)')

    with pytest.raises(TypeError, match="assigned 'info_item'"):
        StreamReaderInfo.continuous('a')


def test_stream_info_without_file_info(monkeypatch):
    """Test that stream info is possible without file info.

    This is the case for scintillometry.io.hdf5.  We use a hack where
    we replace the regular VDIF file reader with an info-less one.
    """
    class NoFileInfoReader(vdif.base.VDIFFileReader):
        @property
        def info(self):
            raise AttributeError

    monkeypatch.setattr(vdif.base, 'VDIFFileReader', NoFileInfoReader)
    with vdif.open(data.SAMPLE_VDIF) as fh:
        info = fh.info
        assert info.file_info is None
        assert info.format == 'vdif'



# Licensed under the GPLv3 - see LICENSE
"""
Definitions for VLBI Mark 4 payloads.

Implements a Mark4Payload class used to store payload words, and decode to
or encode from a data array.

For the specification, see
https://www.haystack.mit.edu/tech/vlbi/mark5/docs/230.3.pdf
"""
import sys
from collections import namedtuple

import numpy as np

from ..base.payload import PayloadBase
from ..base.encoding import encode_2bit_base, decoder_levels
from ..base.utils import fixedvalue
from .header import MARK4_DTYPES


__all__ = ['reorder32', 'reorder64', 'init_luts', 'decode_8chan_2bit_fanout4',
           'encode_8chan_2bit_fanout4', 'Mark4Payload']

#  2bit/fanout4 use the following in decoding 32 and 64 track data:
if sys.byteorder == 'big':  # pragma: no cover
    def reorder32(x):
        """Reorder 32-track bits to bring signs & magnitudes together."""
        return (((x & 0x55AA55AA))
                | ((x & 0xAA00AA00) >> 9)
                | ((x & 0x00550055) << 9))

    def reorder64(x):
        """Reorder 64-track bits to bring signs & magnitudes together."""
        return (((x & 0x55AA55AA55AA55AA))
                | ((x & 0xAA00AA00AA00AA00) >> 9)
                | ((x & 0x0055005500550055) << 9))

    def reorder64_Ft(x):
        """Reorder 64-track bits to bring signs & magnitudes together.

        Special version for the Ft station, which has unusual settings.
        """
        return (((x & 0xAFFAFFFFAFFAFFFF))
                | ((x & 0x0005000000050000) << 12)
                | ((x & 0x5000000050000000) >> 12))
else:
    def reorder32(x):
        """Reorder 32-track bits to bring signs & magnitudes together."""
        return (((x & 0xAA55AA55))
                | ((x & 0x55005500) >> 7)
                | ((x & 0x00AA00AA) << 7))

    # Can speed this up from 140 to 132 us by predefining bit patterns as
    # array scalars.  Inplace calculations do not seem to help much.
    def reorder64(x):
        """Reorder 64-track bits to bring signs & magnitudes together."""
        return (((x & 0xAA55AA55AA55AA55))
                | ((x & 0x5500550055005500) >> 7)
                | ((x & 0x00AA00AA00AA00AA) << 7))

    def reorder64_Ft(x):
        """Reorder 64-track bits to bring signs & magnitudes together.

        Special version for the Ft station, which has unusual settings.
        """
        return (((x & 0xFFFFFAAFFFFFFAAF))
                | ((x & 0x0000050000000500) >> 4)
                | ((x & 0x0000005000000050) << 4))
    # Check on 2015-JUL-12: C code: 738811025863578102 -> 738829572664316278
    # 118, 209, 53, 244, 148, 217, 64, 10
    # reorder64(np.array([738811025863578102], dtype=np.uint64))
    # # array([738829572664316278], dtype=uint64)
    # reorder64(np.array([738811025863578102], dtype=np.uint64)).view(np.uint8)
    # # array([118, 209,  53, 244, 148, 217,  64,  10], dtype=uint8)
    # decode_2bit_64track_fanout4(
    #     np.array([738811025863578102], dtype=np.int64)).astype(int).T
    # -1  1  3  1  array([[-1,  1,  3,  1],
    #  1  1  3 -3         [ 1,  1,  3, -3],
    #  1 -3  1  3         [ 1, -3,  1,  3],
    # -3  1  3  3         [-3,  1,  3,  3],
    # -3  1  1 -1         [-3,  1,  1, -1],
    # -3 -3 -3  1         [-3, -3, -3,  1],
    #  1 -1  1  3         [ 1, -1,  1,  3],
    # -1 -1 -3 -3         [-1, -1, -3, -3]])


def init_luts():
    """Set up the look-up tables for levels as a function of input byte."""
    # Organisation by bits is quite odd for Mark 4.
    b = np.arange(256)[:, np.newaxis]
    # lut1bit
    i = np.arange(8)
    # For all 1-bit modes; if set, sign=-1, so need to get item 0.
    lut1bit = decoder_levels[1][((b >> i) & 1) ^ 1]
    i = np.arange(4)
    # fanout 1 @ 8/16t, fanout 4 @ 32/64t
    s = i*2  # 0, 2, 4, 6
    m = s+1  # 1, 3, 5, 7
    lut2bit1 = decoder_levels[2][2*(b >> s & 1)
                                 + (b >> m & 1)]
    # fanout 2 @ 8/16t, fanout 1 @ 32/64t
    s = i + (i//2)*2  # 0, 1, 4, 5
    m = s + 2         # 2, 3, 6, 7
    lut2bit2 = decoder_levels[2][2*(b >> s & 1)
                                 + (b >> m & 1)]
    # fanout 4 @ 8/16t, fanout 2 @ 32/64t
    s = i    # 0, 1, 2, 3
    m = s+4  # 4, 5, 6, 7
    lut2bit3 = decoder_levels[2][2*(b >> s & 1)
                                 + (b >> m & 1)]
    return lut1bit, lut2bit1, lut2bit2, lut2bit3


lut1bit, lut2bit1, lut2bit2, lut2bit3 = init_luts()

# Look-up table for the number of bits in a byte.
nbits = ((np.arange(256)[:, np.newaxis] >> np.arange(8) & 1)
         .sum(1).astype(np.int16))


def decode_2chan_2bit_fanout4(frame):
    """Decode payload for 2 channels using 2 bits, fan-out 4 (16 tracks)."""
    # header['magnitude_bit'] = 00001111,00001111
    # makes sense with lut2bit3
    # header['fan_out'] = 01230123,01230123
    # header['converter_id'] = 00000000,11111111
    # header['lsb_output'] = 11111111,11111111
    # After reshape: byte 0: ch0/s0, ch0/s1, ch0/s2, ch0/s3, + mag.
    #                byte 1: ch1/s0, ch1/s1, ch1/s2, ch1/s3, + mag.
    frame = frame.view(np.uint8).reshape(-1, 2)
    # The look-up table splits each data word into the above 8 measurements,
    # the transpose pushes channels first and fanout last, and the reshape
    # flattens the fanout.
    return lut2bit3.take(frame, axis=0).transpose(1, 0, 2).reshape(2, -1).T


def encode_2chan_2bit_fanout4(values):
    """Encode payload for 2 channels using 2 bits, fan-out 4 (16 tracks)."""
    # Reverse reshaping (see above).
    values = values.reshape(-1, 4, 2).transpose(0, 2, 1)
    bitvalues = encode_2bit_base(values)
    # Values are -3, -1, +1, 3 -> 00, 01, 10, 11; get first bit (sign) as 1,
    # second bit (magnitude) as 16.
    reorder_bits = np.array([0, 16, 1, 17], dtype=np.uint8)
    reorder_bits.take(bitvalues, out=bitvalues)
    bitvalues <<= np.array([0, 1, 2, 3], dtype=np.uint8)
    out = np.bitwise_or.reduce(bitvalues, axis=-1).ravel().view('<u2')
    return out


def decode_4chan_2bit_fanout4(frame):
    """Decode payload for 4 channels using 2 bits, fan-out 4 (32 tracks)."""
    # Bitwise reordering of tracks, to align sign and magnitude bits,
    # reshaping to get VLBI channels in sequential, but wrong order.
    frame = reorder32(frame.view(np.uint32)).view(np.uint8).reshape(-1, 4)
    # Correct ordering.
    frame = frame.take(np.array([0, 2, 1, 3]), axis=1)
    # The look-up table splits each data byte into 4 measurements.
    # Using transpose ensures channels are first, then time samples, then
    # those 4 measurements, so the reshape orders the samples correctly.
    # Another transpose ensures samples are the first dimension.
    return lut2bit1.take(frame.T, axis=0).reshape(4, -1).T


def encode_4chan_2bit_fanout4(values):
    """Encode payload for 4 channels using 2 bits, fan-out 4 (32 tracks)."""
    reorder_channels = np.array([0, 2, 1, 3])
    values = values[:, reorder_channels].reshape(-1, 4, 4).transpose(0, 2, 1)
    bitvalues = encode_2bit_base(values)
    reorder_bits = np.array([0, 2, 1, 3], dtype=np.uint8)
    reorder_bits.take(bitvalues, out=bitvalues)
    bitvalues <<= np.array([0, 2, 4, 6], dtype=np.uint8)
    out = np.bitwise_or.reduce(bitvalues, axis=-1).ravel().view(np.uint32)
    return reorder32(out).view('<u4')


def decode_8chan_2bit_fanout2(frame):
    """Decode payload for 8 channels using 2 bits, fan-out 4 (32 tracks)."""
    # header['magnitude_bit'] = 00001111,00001111,00001111,00001111
    # makes sense with lut2bit3
    # header['fan_out'] = 00110011,00110011,00110011,00110011
    # i.e., s0s0,s1s1,m0m0,m1m1 for each byte
    # header['converter_id'] = 02020202,13131313,02020202,13131313
    # header['lsb_output'] =   00000000,00000000,11111111,11111111
    # After reshape: byte 0: ch0/s0, ch4/s0, ch0/s1, ch4/s1, + mag.
    #                byte 1: ch1/s0, ch5/s0, ch1/s1, ch5/s1, + mag.
    #                byte 2: ch2/s0, ch6/s0, ch2/s1, ch6/s1, + mag.
    #                byte 3: ch3/s0, ch7/s0, ch3/s1, ch7/s1, + mag.
    frame = frame.view(np.uint8).reshape(-1, 4)
    # The look-up table splits each data word into the above 16 measurements.
    # the first reshape means one gets time, channel&0x3, sample, channel&0x4
    # the transpose makes this channel&0x4, channel&0x3, time, sample.
    # the second reshape (which makes a copy) gets one just channel, time,
    # and the final transpose time, channel.
    return (lut2bit3.take(frame, axis=0).reshape(-1, 4, 2, 2)
            .transpose(3, 1, 0, 2).reshape(8, -1).T)


def encode_8chan_2bit_fanout2(values):
    """Encode payload for 8 channels using 2 bits, fan-out 2 (32 tracks)."""
    # words encode 16 values (see above) in order ch&0x3, sample, ch&0x4
    values = (values.reshape(-1, 2, 2, 4).transpose(0, 3, 1, 2)
              .reshape(-1, 4, 4))
    bitvalues = encode_2bit_base(values)
    # values are -3, -1, +1, 3 -> 00, 01, 10, 11;
    # get first bit (sign) as 1, second bit (magnitude) as 16
    reorder_bits = np.array([0, 16, 1, 17], dtype=np.uint8)
    reorder_bits.take(bitvalues, out=bitvalues)
    bitvalues <<= np.array([0, 1, 2, 3], dtype=np.uint8)
    out = np.bitwise_or.reduce(bitvalues, axis=-1).ravel().view('<u4')
    return out


def decode_16chan_2bit_fanout2_ft(frame):
    """Decode payload for 16 channels using 2 bits, fan-out 2 (64 tracks)."""
    # These Fortaleza files have an unusual ordering:
    # header['magnitude_bit'] = 00000101,10101111,00001111,00001111 * 2
    # White later two are as for lut2bit3, the first two are different.
    # header['fan_out'] = 00110011 * 8
    # i.e., s0s0,s1s1,m0m0,m1m1 for the later bytes.
    # header['converter_id'] = 03030303,04040404,15151515,26262626,
    #                          7a7a7a7a,7b7b7b7b,8c8c8c8c,9d9d9d9d
    # 0, 7 are doubled; those have lsb & usb;
    # header['lsb_output'] =   00001010,00001010,00000000,00000000 * 2
    # Should take magnitude bit literally,
    # byte 0: 0u/s0, 3u/s0, 0u/s1, 3u/s1, 0l/s0, 3u/m0, 0l/s1, 3u/m1
    # byte 1: 0u/m0, 4u/s0, 0u/m1, 4u/s1, 0l/m0, 4u/m0, 0l/m1, 4u/m1
    # byte 2: 1u/s0, 5u/s0, 1u/s1, 5u/s1, 1u/m0, 5u/m0, 1u/m1, 5u/m1
    # byte 3: 2u/s0, 6u/s0, 2u/s1, 6u/s1, 2u/m0, 6u/m0, 2u/m1, 6u/m1
    # byte 4: 7u/s0, au/s0, 7u/s1, au/s1, 7l/s0, au/m0, 7l/s1, au/m1
    # byte 5: 7u/m0, bu/s0, 7u/m1, bu/s1, 7l/m0, bu/m0, 7l/m1, bu/m1
    # byte 6: 8u/s0, cu/s0, 8u/s1, cu/s1, 8u/m0, cu/m0, 8u/m1, cu/m1
    # byte 7: 9u/s0, du/s0, 9u/s1, du/s1, 9u/m0, du/m0, 9u/m1, du/m1
    # This means the re-ordering is different from the usual: we just
    # need to shift bits around in bytes 0,1,4,5:
    frame = reorder64_Ft(frame.view(np.uint64))
    # This leaves samples as
    # byte 0: 0u/s0, 3u/s0, 0u/s1, 3u/s1, 0u/m0, 3u/m0, 0u/m1, 3u/m1
    # byte 1: 0l/s0, 4u/s0, 0l/s1, 4u/s1, 0l/m0, 4u/m0, 0l/m1, 4u/m1
    # byte 2: 1u/s0, 5u/s0, 1u/s1, 5u/s1, 1u/m0, 5u/m0, 1u/m1, 5u/m1
    # byte 3: 2u/s0, 6u/s0, 2u/s1, 6u/s1, 2u/m0, 6u/m0, 2u/m1, 6u/m1
    # byte 4: 7u/s0, au/s0, 7u/s1, au/s1, 7u/m0, au/m0, 7u/m1, au/m1
    # byte 5: 7l/s0, bu/s0, 7l/s1, bu/s1, 7l/m0, bu/m0, 7l/m1, bu/m1
    # byte 6: 8u/s0, cu/s0, 8u/s1, cu/s1, 8u/m0, cu/m0, 8u/m1, cu/m1
    # byte 7: 9u/s0, du/s0, 9u/s1, du/s1, 9u/m0, du/m0, 9u/m1, du/m1
    frame = frame.view(np.uint8).reshape(-1, 8)
    # The look-up table splits each data word into the above 32 measurements.
    # Translating 0u=0, 0l=1, 1..6=2..7, 7u=8, 7l=9, 8..d=a..f, the

    # first reshape yieds time, channel&0x8, channel&0x3, sample, channel&0x4
    # transpose makes this channel&0x8, channel&0x4, channel&0x3, time, sample.
    # and the second reshape (which makes a copy) gets one just time, channel,
    # and the final transpose time, channel.
    return (lut2bit3.take(frame, axis=0).reshape(-1, 2, 4, 2, 2)
            .transpose(1, 4, 2, 0, 3).reshape(16, -1).T)


def encode_16chan_2bit_fanout2_ft(values):
    """Encode payload for 16 channels using 2 bits, fan-out 2 (64 tracks)."""
    # words encode 32 values (above) in order ch&0x8, ch&0x3, sample, ch&0x4
    # First reshape goes to time, sample, ch&0x8, ch&0x4, ch&0x3,
    # transpose makes this time, ch&0x8, ch&0x3, sample, ch&0x4.
    # second reshape future bytes, sample+ch&0x4.
    values = (values.reshape(-1, 2, 2, 2, 4).transpose(0, 2, 4, 1, 3)
              .reshape(-1, 4))
    bitvalues = encode_2bit_base(values)
    # values are -3, -1, +1, 3 -> 00, 01, 10, 11;
    # get first bit (sign) as 1, second bit (magnitude) as 16
    reorder_bits = np.array([0, 16, 1, 17], dtype=np.uint8)
    reorder_bits.take(bitvalues, out=bitvalues)
    bitvalues <<= np.array([0, 1, 2, 3], dtype=np.uint8)
    out = np.bitwise_or.reduce(bitvalues, axis=-1).ravel().view(np.uint64)
    return reorder64_Ft(out).view('<u8')


def decode_8chan_2bit_fanout4(frame):
    """Decode payload for 8 channels using 2 bits, fan-out 4 (64 tracks)."""
    # Bitwise reordering of tracks, to align sign and magnitude bits,
    # reshaping to get VLBI channels in sequential, but wrong order.
    frame = reorder64(frame.view(np.uint64)).view(np.uint8).reshape(-1, 8)
    # Correct ordering.
    frame = frame.take(np.array([0, 2, 1, 3, 4, 6, 5, 7]), axis=1)
    # The look-up table splits each data byte into 4 measurements.
    # Using transpose ensures channels are first, then time samples, then
    # those 4 measurements, so the reshape orders the samples correctly.
    # Another transpose ensures samples are the first dimension.
    return lut2bit1.take(frame.T, axis=0).reshape(8, -1).T


def encode_8chan_2bit_fanout4(values):
    """Encode payload for 8 channels using 2 bits, fan-out 4 (64 tracks)."""
    reorder_channels = np.array([0, 2, 1, 3, 4, 6, 5, 7])
    values = values[:, reorder_channels].reshape(-1, 4, 8).transpose(0, 2, 1)
    bitvalues = encode_2bit_base(values)
    reorder_bits = np.array([0, 2, 1, 3], dtype=np.uint8)
    reorder_bits.take(bitvalues, out=bitvalues)
    bitvalues <<= np.array([0, 2, 4, 6], dtype=np.uint8)
    out = np.bitwise_or.reduce(bitvalues, axis=-1).ravel().view(np.uint64)
    return reorder64(out).view('<u8')


class Mark4Payload(PayloadBase):
    """Container for decoding and encoding Mark 4 payloads.

    Parameters
    ----------
    words : `~numpy.ndarray`
        Array containg LSB unsigned words (with the right size) that
        encode the payload.
    header : `~baseband.mark4.Mark4Header`, optional
        If given, used to infer the number of channels, bps, and fanout.
    sample_shape : tuple
        Shape of the samples (e.g., (nchan,)).  Default: (1,).
    bps : int, optional
        Number of bits per sample, used if ``header`` is not given.
        Default: 2.
    fanout : int, optional
        Number of tracks every bit stream is spread over, used if ``header`` is
        not given.  Default: 1.
    magnitude_bit : int, optional
        Magnitude bits for all tracks packed together. Used to index
        encoder and decoder.  Default: assume standard Mark 4 payload,
        for which number of channels, bps, and fanout suffice.

    Notes
    -----
    The total number of tracks is ``nchan * bps * fanout``.
    """

    _dtype_word = None
    # Decoders keyed by (nchan, nbit, fanout).
    _encoders = {(2, 2, 4): encode_2chan_2bit_fanout4,
                 (4, 2, 4): encode_4chan_2bit_fanout4,
                 (8, 2, 2): encode_8chan_2bit_fanout2,
                 (8, 2, 4): encode_8chan_2bit_fanout4,
                 (16, 0xf0faf050f0faf05, 2): encode_16chan_2bit_fanout2_ft}
    _decoders = {(2, 2, 4): decode_2chan_2bit_fanout4,
                 (4, 2, 4): decode_4chan_2bit_fanout4,
                 (8, 2, 2): decode_8chan_2bit_fanout2,
                 (8, 2, 4): decode_8chan_2bit_fanout4,
                 (16, 0xf0faf050f0faf05, 2): decode_16chan_2bit_fanout2_ft}

    _sample_shape_maker = namedtuple('SampleShape', 'nchan')

    def __init__(self, words, header=None, *, sample_shape=(1,), bps=2,
                 fanout=1, magnitude_bit=None, complex_data=False):
        if header is not None:
            magnitude_bit = header['magnitude_bit']
            bps = 2 if magnitude_bit.any() else 1
            ta = header.track_assignment
            if bps == 1 or np.all(magnitude_bit[ta] == [False, True]):
                magnitude_bit = None
            else:
                magnitude_bit = (np.packbits(magnitude_bit)
                                 .view(header.stream_dtype).item())

            ntrack = header.ntrack
            fanout = header.fanout
            sample_shape = (ntrack // (bps * fanout),)
            self._nbytes = header.payload_nbytes
        else:
            ntrack = sample_shape[0] * bps * fanout
            magnitude_bit = None

        self._dtype_word = MARK4_DTYPES[ntrack]
        self.fanout = fanout
        super().__init__(words, sample_shape=sample_shape,
                         bps=bps, complex_data=complex_data)
        self._coder = (self.sample_shape.nchan,
                       (self.bps if magnitude_bit is None else magnitude_bit),
                       self.fanout)

    @fixedvalue
    def complex_data(self):
        return False

    @classmethod
    def fromfile(cls, fh, header=None, **kwargs):
        """Read payload from filehandle and decode it into data.

        Parameters
        ----------
        fh : filehandle
            From which data is read.
        header : `~baseband.mark4.Mark4Header`
            Used to infer ``payload_nbytes``, ``bps``, ``sample_shape``, and
            ``dtype``.  If not given, those have to be passed in.
        """
        if header is not None:
            kwargs.setdefault('dtype', header.stream_dtype)
        return super().fromfile(fh, header=header, **kwargs)

    @classmethod
    def fromdata(cls, data, header):
        """Encode data as payload, using header information."""
        if data.dtype.kind == 'c':
            raise ValueError("Mark4 format does not support complex data.")
        if header.sample_shape != data.shape[1:]:
            raise ValueError("header is for {0} channels but data has {1}"
                             .format(header.nchan, data.shape[-1]))
        words = np.empty(header.payload_nbytes // header.stream_dtype.itemsize,
                         header.stream_dtype)
        self = cls(words, header)
        self[:] = data
        return self



# Licensed under the GPLv3 - see LICENSE.rst
import os
import pickle

import pytest
import numpy as np

from .. import sequentialfile as sf

from baseband import vdif
from baseband.data import SAMPLE_VDIF as SAMPLE_FILE


class Sequencer:
    def __init__(self, template):
        self.template = template

    def __getitem__(self, item):
        return self.template.format(item)


class TestSequentialFileReader:

    def _setup(self, tmpdir):
        self.data = b'abcdefghijklmnopqrstuvwxyz'
        self.uint8_data = np.frombuffer(self.data, dtype=np.uint8)
        self.size = len(self.data)
        self.files = [str(tmpdir.join('file{:1d}.raw'.format(i)))
                      for i in range(3)]
        self.max_file_size = 10
        self.sizes = []
        self.offsets = [0]
        offset = 0
        for filename in self.files:
            with open(filename, 'wb') as fw:
                part = self.data[offset:offset+self.max_file_size]
                fw.write(part)
                self.sizes.append(len(part))
                self.offsets.append(self.offsets[-1] + len(part))
            offset += self.max_file_size

    def test_setup(self, tmpdir):
        self._setup(tmpdir)
        assert self.sizes == [10, 10, 6]
        assert self.offsets == [0, 10, 20, 26]

    def test_open_close(self, tmpdir):
        self._setup(tmpdir)

        fh = sf.open(self.files)
        assert fh.readable()
        assert fh.seekable()
        assert not fh.writable()
        assert not fh.closed
        assert repr(fh).startswith('SequentialFileReader(files=')
        fh.close()
        assert fh.closed
        with pytest.raises(ValueError):
            # wrong mode: no r or w
            sf.open(self.files, 'b')
        with pytest.raises(TypeError):
            # cannot pass in file_size for reading
            sf.open(self.files, 'rb', file_size=10)

    def test_context(self, tmpdir):
        self._setup(tmpdir)
        with sf.open(self.files) as fh:
            assert fh.readable()
            assert fh.seekable()
            assert not fh.writable()
            assert not fh.closed
        assert fh.closed

    def test_attributes(self, tmpdir):
        self._setup(tmpdir)
        with sf.open(self.files) as fh:
            assert fh.files == self.files
            assert fh.tell() == 0
            assert fh.size == self.size
            assert fh._file_sizes == self.sizes
            assert fh._file_offsets == self.offsets
            with pytest.raises(AttributeError):
                fh.bla

    def test_seek(self, tmpdir):
        self._setup(tmpdir)

        with sf.open(self.files) as fh:
            fh.seek(self.offsets[1])
            assert fh.file_nr == 1
            assert fh.tell() == self.offsets[1]
            fh.seek(self.offsets[1] + 1)
            assert fh.file_nr == 1
            fh.seek(self.offsets[2])
            assert fh.file_nr == 2
            fh.seek(self.offsets[3])
            assert fh.tell() == self.size
            assert fh.file_nr == 2
            fh.seek(self.size)
            assert fh.file_nr == 2
            fh.seek(10, 2)
            assert fh.file_nr == 2
            assert fh.tell() == self.size + 10
            with pytest.raises(OSError):
                fh.seek(-3)
            fh.seek(-3, 2)
            assert fh.tell() == self.size - 3
            assert fh.file_nr == 2
            fh.seek(-5, 1)
            assert fh.tell() == self.size - 3 - 5
            assert fh.file_nr == 1
            with pytest.raises(ValueError):
                fh.seek(-5, -1)
            with pytest.raises(ValueError):
                fh.seek(-5, 3)
        # cannot seek closed file
        with pytest.raises(ValueError):
            fh.seek(10)

    def test_read(self, tmpdir):
        self._setup(tmpdir)

        with sf.open(self.files) as fh:
            check = fh.read(2)
            assert check == self.data[:2]
            check = fh.read(2)
            assert check == self.data[2:4]
            check = fh.read(10)
            assert check == self.data[4:14]
            check = fh.read()
            assert check == self.data[14:]
            fh.seek(0)
        # cannot read closed file
        with pytest.raises(ValueError):
            fh.read()

    def test_read_all(self, tmpdir):
        self._setup(tmpdir)
        with sf.open(self.files) as fh:
            check = fh.read()
            assert check == self.data

    def test_seek_read(self, tmpdir):
        self._setup(tmpdir)
        with sf.open(self.files) as fh:
            fh.seek(8)
            assert fh.read(8) == self.data[8:16]
            fh.seek(1, 1)
            assert fh.read(8) == self.data[17:25]
            fh.seek(-2, 1)
            assert fh.read(2) == self.data[23:25]
            fh.seek(-10, 2)
            assert fh.read() == self.data[-10:]
            fh.seek(10, 2)
            assert fh.read() == b''
            assert fh.read(10) == b''

    def test_memmap(self, tmpdir):
        self._setup(tmpdir)

        with sf.open(self.files) as fh:
            mm = fh.memmap(offset=0, shape=(5,))
            assert fh.tell() == 5
            assert (mm == self.uint8_data[:5]).all()
            mm = fh.memmap(shape=(5,))
            assert fh.tell() == 10
            assert (mm == self.uint8_data[5:10]).all()
            with pytest.raises(ValueError):
                fh.memmap(offset=7, shape=(5,))
            offset = self.offsets[1]
            fh.seek(offset)
            mm = fh.memmap(shape=5)
            assert (mm == self.uint8_data[offset:offset+5]).all()
            fh.seek(-2, 2)
            mm = fh.memmap()
            assert (mm == self.uint8_data[-2:]).all()
            fh.seek(-4, 2)
            mm = fh.memmap(mode='r', dtype=np.uint16)
            assert (mm == self.uint8_data[-4:].view(np.uint16)).all()
            fh.seek(-3, 2)
            with pytest.raises(ValueError):
                fh.memmap(dtype=np.uint16)

        with pytest.raises(ValueError):  # file closed.
            fh.memmap(offset=0, shape=(5,))

    @pytest.mark.parametrize('action', [None, 'seek', 'read'])
    def test_pickle(self, tmpdir, action):
        self._setup(tmpdir)
        with sf.open(self.files) as fh:
            if action == 'seek':
                fh.seek(16)
            elif action == 'read':
                fh.seek(8)
                fh.read(8)
            pickled = pickle.dumps(fh)
            with pickle.loads(pickled) as fh2:
                if action is None:
                    assert fh2.read(16) == self.data[:16]
                assert fh2.tell() == 16
                assert fh2.read(9) == self.data[16:25]
                fh2.seek(-2, 1)
                assert fh2.read(2) == self.data[23:25]
                fh2.seek(0)
                assert fh2.read() == self.data

            # cannot read closed file
            with pytest.raises(ValueError):
                fh2.read()

            if action:
                assert fh.tell() == 16
                assert fh.read(1) == self.data[16:17]
            else:
                assert fh.tell() == 0
                assert fh.read(1) == self.data[:1]

        with pickle.loads(pickled) as fh3:
            if action is None:
                fh3.seek(16)
            else:
                assert fh3.tell() == 16
            assert fh3.read(9) == self.data[16:25]

        # cannot read closed file
        with pytest.raises(ValueError):
            fh3.read()

        closed = pickle.dumps(fh)
        with pickle.loads(closed) as fh4:
            assert fh4.closed
            with pytest.raises(ValueError):
                fh4.read()


class TestSequentialFileWriter:
    def _setup(self, tmpdir):
        self.data = b'abcdefghijklmnopqrstuvwxyz'
        self.uint8_data = np.frombuffer(self.data, dtype=np.uint8)
        self.files = [str(tmpdir.join('file{:1d}.raw'.format(i)))
                      for i in range(3)]

    def files_exist(self):
        return [os.path.isfile(fil) for fil in self.files]

    def test_open_close(self, tmpdir):
        self._setup(tmpdir)
        fh = sf.open(self.files, 'wb', file_size=10)
        assert fh.writable()
        assert not fh.closed
        fh.close()
        assert fh.closed
        assert self.files_exist() == [True, False, False]

    def test_context(self, tmpdir):
        self._setup(tmpdir)
        with sf.open(self.files, 'wb', file_size=10) as fh:
            assert fh.writable()
            assert not fh.closed

        assert fh.closed
        assert self.files_exist() == [True, False, False]

    def test_attributes(self, tmpdir):
        self._setup(tmpdir)
        with sf.open(self.files, 'wb', file_size=10) as fh:
            assert fh.files == self.files
            assert fh.file_size == 10
            assert fh.tell() == 0

    def test_write(self, tmpdir):
        self._setup(tmpdir)
        with sf.open(self.files, 'wb', file_size=10) as fh:
            fh.write(self.data[:5])
            assert fh.tell() == 5
            assert fh.file_nr == 0
            assert self.files_exist() == [True, False, False]
            fh.write(self.data[5:20])
            assert fh.tell() == 20
            assert fh.file_nr == 1
            assert self.files_exist() == [True, True, False]
            fh.write(self.data[20:])
            assert fh.tell() == len(self.data)
            assert fh.file_nr == 2

        # Cannot write to closed file.
        with pytest.raises(ValueError):
            fh.write(b' ')

        assert self.files_exist() == [True, True, True]
        with sf.open(self.files, 'rb') as fh:
            assert fh.read() == self.data

        with pytest.raises(OSError):
            with sf.open(self.files, 'wb', file_size=10) as fh:
                fh.write(b' ' * (len(self.files) * 10 + 1))

    def test_write_all(self, tmpdir):
        self._setup(tmpdir)
        with sf.open(self.files, 'wb', file_size=10) as fh:
            fh.write(self.data)
            assert fh.tell() == len(self.data)
            assert fh.file_nr == 2

        assert self.files_exist() == [True, True, True]
        with sf.open(self.files, 'rb') as fh:
            assert fh.read() == self.data

    def test_simple_sequencer(self, tmpdir):
        self._setup(tmpdir)

        sequencer = Sequencer(str(tmpdir.join('file{:03d}.raw')))
        with sf.open(sequencer, 'wb', file_size=8) as fh:
            fh.write(self.data)
            assert fh.file_nr == 3

        with sf.open(sequencer, 'rb') as fh:
            check = fh.read()
            assert check == self.data
            assert fh._file_sizes == [8, 8, 8, 2]

    def test_memmap(self, tmpdir):
        self._setup(tmpdir)
        data = self.uint8_data
        with sf.open(self.files, 'w+b', file_size=10) as fh:
            mm = fh.memmap(shape=(5,))
            assert fh.tell() == 5
            mm[:] = data[:5]
            mm2 = fh.memmap(offset=5, shape=(5,))
            mm2[:4] = data[5:9]
            mm2[-1] = data[-1]
            mm3 = fh.memmap(shape=(5,))
            mm3[:] = data[10:15]
            with pytest.raises(ValueError):
                fh.memmap(shape=(6,))  # beyond file.
            with pytest.raises(ValueError):
                fh.memmap()  # Need to pass in shape
            with pytest.raises(ValueError):
                fh.memmap(offset=4)  # Cannot seek
            mm4 = fh.memmap(dtype=np.uint16, shape=(2,))
            mm4[:] = data[15:19].view(np.uint16)

        assert self.files_exist() == [True, True, False]
        with sf.open(self.files[:2], 'rb') as fh:
            check = fh.read()
        assert len(check) == 19
        assert check[:9] == self.data[:9]
        assert check[9] == self.data[-1]
        assert check[10:] == self.data[10:19]

    def test_memmap_fail(self, tmpdir):
        self._setup(tmpdir)
        with sf.open(self.files, 'wb', file_size=10) as fh:
            with pytest.raises(ValueError):
                fh.memmap(shape=(5,))

    def test_pickle_fails(self, tmpdir):
        self._setup(tmpdir)
        with sf.open(self.files, 'wb', file_size=10) as fh:
            with pytest.raises(TypeError):
                pickle.dumps(fh)


class TestFileNameSequencer:
    def setup(self):
        with open(SAMPLE_FILE, 'rb') as fh:
            self.header = vdif.VDIFHeader.fromfile(fh)

    def test_enumeration(self):
        fns1 = sf.FileNameSequencer('x{file_nr:03d}.vdif')
        assert fns1[0] == 'x000.vdif'
        assert fns1[107] == 'x107.vdif'
        fns2 = sf.FileNameSequencer('{SNAKE}_{file_nr}', {'SNAKE': 'python'})
        assert fns2[13] == 'python_13'

        with pytest.raises(KeyError):
            sf.FileNameSequencer('{snake:06d}.x', {'SNAKE': 10})

    def test_header_extraction(self):
        template = 'x.edv{edv}.stn_{station_id}.{file_nr:05d}.vdif'
        fns = sf.FileNameSequencer(template, self.header)
        assert fns[0] == 'x.edv3.stn_65532.00000.vdif'
        assert fns[133] == 'x.edv3.stn_65532.00133.vdif'

    def test_len(self, tmpdir):
        template = str(tmpdir.join('a{file_nr}.bin'))
        fns = sf.FileNameSequencer(template)
        for i in range(5):
            assert len(fns) == i
            filename = fns[i]
            assert filename.endswith('a{}.bin'.format(i))
            with open(filename, 'wb') as fh:
                fh.write(b'bird')
        assert len(fns) == 5
        assert fns[-2] == fns[3]
        assert fns[-1].endswith('a4.bin')
        with pytest.raises(IndexError):
            fns[-10]



# Licensed under the GPLv3 - see LICENSE
#
# __      __  _____    _   _____
# \ \    / / | ___ \  | | |   __|
#  \ \  / /  | |  | | | | |  |_
#   \ \/ /   | |  | | | | |   _]
#    \  /    | |__| | | | |  |
#     \/     |_____/  |_| |__|
#
#
"""VLBI Data Interchange Format (VDIF) reader/writer

For the VDIF specification, see https://vlbi.org/vlbi-standards/vdif/
"""
from .base import open, info  # noqa
from .header import VDIFHeader  # noqa
from .payload import VDIFPayload  # noqa
from .frame import VDIFFrame, VDIFFrameSet  # noqa



# Licensed under the GPLv3 - see LICENSE
import io
import re
import operator
from functools import reduce

import astropy.units as u
from astropy.utils import lazyproperty

from ..helpers import sequentialfile as sf
from ..base.base import (
    FileBase,
    StreamReaderBase, StreamWriterBase,
    FileOpener, FileInfo)
from ..base.file_info import FileReaderInfo
from ..base.utils import lcm
from .header import DADAHeader
from .payload import DADAPayload
from .frame import DADAFrame


__all__ = ['DADAFileNameSequencer', 'DADAFileReader', 'DADAFileWriter',
           'DADAStreamBase', 'DADAStreamReader', 'DADAStreamWriter',
           'open', 'info']


class DADAFileNameSequencer(sf.FileNameSequencer):
    """List-like generator of DADA filenames using a template.

    The template is formatted, filling in any items in curly brackets with
    values from the header, as well as possibly a file number equal to the
    indexing value, indicated with '{file_nr}'.  The value '{obs_offset}' is
    treated specially, in being calculated using ``header['OBS_OFFSET'] +
    file_nr * header['FILE_SIZE']``, where ``header['FILE_SIZE']`` is the file
    size in bytes.

    The length of the instance will be the number of files that exist that
    match the template for increasing values of the file number (when writing,
    it is the number of files that have so far been generated).

    Parameters
    ----------
    template : str
        Template to format to get specific filenames.  Curly bracket item
        keywords are not case-sensitive.
    header : dict-like
        Structure holding key'd values that are used to fill in the format.
        Keys must be in all caps (eg. ``DATE``), as with DADA header keys.

    Examples
    --------

    >>> from baseband import dada
    >>> dfs = dada.DADAFileNameSequencer(
    ...     '{date}_{file_nr:03d}.dada', {'DATE': "2018-01-01"})
    >>> dfs[10]
    '2018-01-01_010.dada'
    >>> from baseband.data import SAMPLE_DADA
    >>> with open(SAMPLE_DADA, 'rb') as fh:
    ...     header = dada.DADAHeader.fromfile(fh)
    >>> template = '{utc_start}.{obs_offset:016d}.000000.dada'
    >>> dfs = dada.DADAFileNameSequencer(template, header)
    >>> dfs[0]
    '2013-07-02-01:37:40.0000006400000000.000000.dada'
    >>> dfs[1]
    '2013-07-02-01:37:40.0000006400064000.000000.dada'
    >>> dfs[10]
    '2013-07-02-01:37:40.0000006400640000.000000.dada'
    """

    def __init__(self, template, header={}):
        self.items = {}

        def check_and_convert(x):
            string = x.group().upper()
            key = string[1:-1]
            if key != 'FRAME_NR' and key != 'FILE_NR':
                self.items[key] = header[key]
            return string

        # This converts template names to upper case, since header keywords are
        # all upper case.
        self.template = re.sub(r'{\w+[}:]', check_and_convert, template)
        self._has_obs_offset = 'OBS_OFFSET' in self.items
        if self._has_obs_offset:
            self._obs_offset0 = self.items['OBS_OFFSET']
            self._file_size = header['FILE_SIZE']

    def _process_items(self, file_nr):
        super()._process_items(file_nr)
        # Pop file_nr, as we need to capitalize it.
        file_nr = self.items.pop('file_nr')
        self.items['FRAME_NR'] = self.items['FILE_NR'] = file_nr
        if self._has_obs_offset:
            self.items['OBS_OFFSET'] = (self._obs_offset0
                                        + file_nr * self._file_size)


class DADAFileReader(FileBase):
    """Simple reader for DADA files.

    Wraps a binary filehandle, providing methods to help interpret the data,
    such as `read_frame` and `get_frame_rate`. By default, frame payloads
    are mapped rather than fully read into physical memory.

    Parameters
    ----------
    fh_raw : filehandle
        Filehandle of the raw binary data file.
    """
    info = FileReaderInfo()

    def read_header(self):
        """Read a single header from the file.

        Returns
        -------
        header : `~baseband.dada.DADAHeader`
        """
        return DADAHeader.fromfile(self.fh_raw)

    def read_frame(self, memmap=True, verify=True):
        """Read the frame header and read or map the corresponding payload.

        Parameters
        ----------
        memmap : bool, optional
            If `True` (default), map the payload using `~numpy.memmap`, so that
            parts are only loaded into memory as needed to access data.
        verify : bool, optional
            Whether to do basic checks of frame integrity.  Default: `True`.

        Returns
        -------
        frame : `~baseband.dada.DADAFrame`
            With ``.header`` and ``.payload`` properties.  The ``.data``
            property returns all data encoded in the frame.  Since this may
            be too large to fit in memory, it may be better to access the
            parts of interest by slicing the frame.
        """
        return DADAFrame.fromfile(self.fh_raw, memmap=memmap, verify=verify)

    def get_frame_rate(self):
        """Determine the number of frames per second.

        The routine uses the sample rate and number of samples per frame
        from the first header in the file.

        Returns
        -------
        frame_rate : `~astropy.units.Quantity`
            Frames per second.
        """
        with self.temporary_offset(0):
            header = self.read_header()
        return (header.sample_rate / header.samples_per_frame).to(u.Hz)


class DADAFileWriter(FileBase):
    """Simple writer/mapper for DADA files.

    Adds `write_frame` and `memmap_frame` methods to the binary file wrapper.
    The latter allows one to encode data in pieces, writing to disk as needed.
    """

    def write_frame(self, data, header=None, **kwargs):
        """Write a single frame (header plus payload).

        Parameters
        ----------
        data : `~numpy.ndarray` or `~baseband.dada.DADAFrame`
            If an array, a ``header`` should be given, which will be used to
            get the information needed to encode the array, and to construct
            the DADA frame.
        header : `~baseband.dada.DADAHeader`
            Can instead give keyword arguments to construct a header.  Ignored
            if ``data`` is a `~baseband.dada.DADAFrame` instance.
        **kwargs
            If ``header`` is not given, these are used to initialize one.
        """
        if not isinstance(data, DADAFrame):
            data = DADAFrame.fromdata(data, header, **kwargs)
        return data.tofile(self.fh_raw)

    def memmap_frame(self, header=None, **kwargs):
        """Get frame by writing the header to disk and mapping its payload.

        The header is written to disk immediately, but the payload is mapped,
        so that it can be filled in pieces, by setting slices of the frame.

        Parameters
        ----------
        header : `~baseband.dada.DADAHeader`
            Written to disk immediately.  Can instead give keyword arguments to
            construct a header.
        **kwargs
            If ``header`` is not given, these are used to initialize one.

        Returns
        -------
        frame: `~baseband.dada.DADAFrame`
            By assigning slices to data, the payload can be encoded piecewise.
        """
        if header is None:
            header = DADAHeader.fromvalues(**kwargs)
        header.tofile(self.fh_raw)
        payload = DADAPayload.fromfile(self.fh_raw, memmap=True, header=header)
        return DADAFrame(header, payload)


class DADAStreamBase:
    """Provides sample shape maker and fast index getting/setting."""

    _sample_shape_maker = DADAPayload._sample_shape_maker

    def _get_index(self, header):
        # Override for faster calculation of frame index.
        return int(round((header['OBS_OFFSET']
                          - self.header0['OBS_OFFSET'])
                         / self.header0.payload_nbytes))

    def _set_index(self, header, index):
        header.update(obs_offset=self.header0['OBS_OFFSET']
                      + index * self.header0.payload_nbytes)


class DADAStreamReader(DADAStreamBase, StreamReaderBase):
    """DADA format reader.

    Allows access to DADA files as a continuous series of samples.

    Parameters
    ----------
    fh_raw : filehandle
        Filehandle of the raw DADA stream.
    squeeze : bool, optional
        If `True` (default), remove any dimensions of length unity from
        decoded data.
    subset : indexing object or tuple of objects, optional
        Specific components of the complete sample to decode (after possibly
        squeezing).  If a single indexing object is passed, it selects
        polarizations.  With a tuple, the first selects polarizations and the
        second selects channels.  If the tuple is empty (default), all
        components are read.
    verify : bool, optional
        Whether to do basic checks of frame integrity when reading.  The first
        frame of the stream is always checked, so ``verify`` is effective only
        when reading sequences of files.  Default: `True`.
    """

    def __init__(self, fh_raw, squeeze=True, subset=(), verify=True):
        fh_raw = DADAFileReader(fh_raw)
        header0 = fh_raw.read_header()
        super().__init__(fh_raw, header0, squeeze=squeeze, subset=subset,
                         verify=verify)
        # Store number of frames, for finding last header.
        with self.fh_raw.temporary_offset() as fh_raw:
            self._raw_file_size = fh_raw.seek(0, 2)
            self._nframes, partial_frame_nbytes = divmod(
                self._raw_file_size, self.header0.frame_nbytes)
            # If there is a partial last frame.
            if partial_frame_nbytes > 0:
                # If partial last frame contains payload bytes.
                if partial_frame_nbytes > self.header0.nbytes:
                    self._nframes += 1
                    # If there's only one frame and it's incomplete.
                    if self._nframes == 1:
                        self._header0 = self._last_header
                        self._samples_per_frame = (
                            self._last_header.samples_per_frame)
                # Otherwise, ignore the partial frame unless it's the only
                # frame, in which case raise an EOFError.
                elif self._nframes == 0:
                    raise EOFError('file (of {0} bytes) appears to end without'
                                   'any payload.'.format(partial_frame_nbytes))

    @lazyproperty
    def _last_header(self):
        """Header of the last file for this stream.

        If last frame is prematurely truncated, header's payload_nbytes is
        reduced accordingly to let the stream reader read to the end of file.
        """
        # We assume DADA files are complete except for possibly the last one.
        # Hence, we seek directly to where the last header should be, but
        # adjust it if the last file is short.
        with self.fh_raw.temporary_offset() as fh_raw:
            self._seek_frame(self._nframes - 1)
            header = fh_raw.read_header()
            payload_nbytes = self._raw_file_size - fh_raw.tell()
            assert payload_nbytes > 0, 'setup failed: no payload in last frame'
            if header.payload_nbytes > payload_nbytes:
                # Truncated last frame.  Adjust header to give the actual
                # number of useful samples, insisting that the payload has
                # integer number of both words and complete samples.
                header.mutable = True
                payload_block = lcm(
                    DADAPayload._dtype_word.itemsize,
                    reduce(operator.mul, self.sample_shape,
                           self.header0.bps * (2 if self.header0.complex_data
                                               else 1) // 8))
                header.payload_nbytes = ((payload_nbytes // payload_block)
                                         * payload_block)
                header.mutable = False

        return header

    @lazyproperty
    def stop_time(self):
        """Time at the end of the file, just after the last sample.

        See also `start_time` for the start time of the file, and `time` for
        the time of the sample pointer's current offset.
        """
        return (self._get_time(self._last_header)
                + (self._last_header.samples_per_frame
                   / self.sample_rate).to(u.s))

    def _fh_raw_read_frame(self):
        # Override to special-case last frame, which may be short.
        if (self.fh_raw.tell() // self.header0.frame_nbytes
                < self._nframes - 1):
            return self.fh_raw.read_frame()

        # Use last header, which will have properly adjusted payload size.
        self.fh_raw.seek(self.header0.nbytes, 1)
        last_payload = DADAPayload.fromfile(self.fh_raw, memmap=True,
                                            header=self._last_header)
        # Ensure we skip all the way to the end of the file, to indicate
        # there is no use in trying to check for the next header.
        self.fh_raw.seek(0, 2)
        return DADAFrame(self._last_header, last_payload)


class DADAStreamWriter(DADAStreamBase, StreamWriterBase):
    """DADA format writer.

    Encodes and writes sequences of samples to file.

    Parameters
    ----------
    raw : filehandle
        For writing the header and raw data to storage.
    header0 : :class:`~baseband.dada.DADAHeader`
        Header for the first frame, holding time information, etc.
    squeeze : bool, optional
        If `True` (default), `write` accepts squeezed arrays as input,
        and adds any dimensions of length unity.
    """

    def __init__(self, fh_raw, header0, squeeze=True):
        assert header0.get('OBS_OVERLAP', 0) == 0
        fh_raw = DADAFileWriter(fh_raw)
        super().__init__(fh_raw, header0, squeeze=squeeze)

    def _make_frame(self, index):
        header = self.header0.copy()
        self._set_index(header, index)
        return self.fh_raw.memmap_frame(header)

    def _fh_raw_write_frame(self, frame):
        assert frame is self._frame
        # Deleting frame flushes memmap'd data to disk.
        del self._frame


class DADAFileOpener(FileOpener):
    FileNameSequencer = DADAFileNameSequencer

    def get_fns(self, name, mode, kwargs):
        fns = super().get_fns(name, mode, kwargs)
        # For obs_offset we need the first file to know the
        # actual file_size.
        if mode[0] == 'r' and 'obs_offset' in name.lower():
            with io.open(fns[0], 'rb') as fh:
                header0 = DADAHeader.fromfile(fh)
            fns = self.FileNameSequencer(name, header0)
        return fns

    def get_fh(self, name, mode, kwargs):
        if mode == 'ws' and self.is_sequence(name):
            kwargs.setdefault('file_size', kwargs['header0'].frame_nbytes)

        return super().get_fh(name, mode, kwargs)


open = DADAFileOpener.create(globals(), doc="""
--- For reading a stream : (see :class:`~baseband.dada.base.DADAStreamReader`)

squeeze : bool, optional
    If `True` (default), remove any dimensions of length unity from
    decoded data.
subset : indexing object or tuple of objects, optional
    Specific components of the complete sample to decode (after possibly
    squeezing).  If a single indexing object is passed, it selects
    polarizations.  With a tuple, the first selects polarizations and the
    second selects channels.  If the tuple is empty (default), all
    components are read.

--- For writing a stream : (see :class:`~baseband.dada.base.DADAStreamWriter`)

header0 : `~baseband.dada.DADAHeader`
    Header for the first frame, holding time information, etc.  Can instead
    give keyword arguments to construct a header (see ``**kwargs``).
squeeze : bool, optional
    If `True` (default), writer accepts squeezed arrays as input, and adds
    any dimensions of length unity.
**kwargs
    If no header is given, an attempt is made to construct one from these.
    For a standard header, this would include the following.

--- Header keywords : (see :meth:`~baseband.dada.DADAHeader.fromvalues`)

time : `~astropy.time.Time`
    Start time of the file.
samples_per_frame : int,
    Number of complete samples per frame.
sample_rate : `~astropy.units.Quantity`
    Number of complete samples per second, i.e. the rate at which each
    channel of each polarization is sampled.
offset : `~astropy.units.Quantity` or `~astropy.time.TimeDelta`, optional
    Time offset from the start of the whole observation (default: 0).
npol : int, optional
    Number of polarizations (default: 1).
nchan : int, optional
    Number of channels (default: 1).
complex_data : bool, optional
    Whether data are complex (default: `False`).
bps : int, optional
    Bits per elementary sample, i.e. per real or imaginary component for
    complex data (default: 8).

Returns
-------
Filehandle
    :class:`~baseband.dada.base.DADAFileReader` or
    :class:`~baseband.dada.base.DADAFileWriter` (binary), or
    :class:`~baseband.dada.base.DADAStreamReader` or
    :class:`~baseband.dada.base.DADAStreamWriter` (stream).

Notes
-----
For streams, one can also pass to ``name`` a list of files, or a template
string that can be formatted using 'frame_nr', 'obs_offset', and other header
keywords (by `~baseband.dada.DADAFileNameSequencer`).

For writing, one can mimic what is done at quite a few telescopes by using
the template '{utc_start}_{obs_offset:016d}.000000.dada'.  Unlike for most
openers, ``file_size`` is set to the size of one frame as given by the header.

For reading, to read series such as the above, use something like
'2013-07-02-01:37:40_{obs_offset:016d}.000000.dada'.  Note that here we
have to pass in the date explicitly, since the template is used to get the
first file name, before any header is read, and therefore the only keywords
available are 'frame_nr', 'file_nr', and 'obs_offset', all of which are
assumed to be zero for the first file. To avoid this restriction, pass in
keyword arguments with values appropriate for the first file.

One may also pass in a `~baseband.helpers.sequentialfile` object
(opened in 'rb' mode for reading or 'w+b' for writing), though for typical use
cases it is practically identical to passing in a list or template.
""")


info = FileInfo.create(globals())



# Licensed under the GPLv3 - see LICENSE
"""Base implementations shared between all formats.

Files are considered as composed of multiple frames, each of which have a
header and payload, which are encoded in various ways. Base classes
implementing the decoding and encoding and exposing a standardized
interface are found in the corresponding `~baseband.base.header`,
`~baseband.base.payload` and `~baseband.base.frame` modules, with the
separate `~baseband.base.encoding` module providing implementations for
common encoding formats.

The `~baseband.base.base` module defines base methods for file and stream
readers and writers that read or write the frames, including possibly
dealing with corrupted data, using the `~baseband.base.offsets` module to
keep track of missing pieces.  Each file and stream reader has an ``info``
property, defined in `~baseband.base.file_info`, that provides standardized
information.

Finally, `~baseband.base.utils` contains some general utility routines such
as for BCD encoding and decoding, and cyclic redundancy checks.

"""



import warnings

from ..base import offsets, utils  # noqa


warnings.warn('the vlbi_base module has been deprecated in baseband 4.0. '
              'Use the base module instead.', DeprecationWarning)



# Licensed under the GPLv3 - see LICENSE
import os
import copy
import pickle

import pytest
import numpy as np
import astropy.units as u
from astropy.time import Time
from ... import dada
from ...helpers import sequentialfile as sf
from ..base import DADAFileNameSequencer
from ...data import SAMPLE_DADA as SAMPLE_FILE


class TestDADA:
    def setup_class(cls):
        with open(SAMPLE_FILE, 'rb') as fh:
            cls.header = dada.DADAHeader.fromfile(fh)
            cls.payload = dada.DADAPayload.fromfile(fh, cls.header,
                                                    memmap=False)

    def test_header(self, tmpdir):
        with open(SAMPLE_FILE, 'rb') as fh:
            header = dada.DADAHeader.fromfile(fh)
            assert header.nbytes == 4096
            assert fh.tell() == 4096
        assert header['NDIM'] == 2
        assert header['NCHAN'] == 1
        assert header['UTC_START'] == '2013-07-02-01:37:40'
        assert header['OBS_OFFSET'] == 6400000000  # 100 s
        assert header.time.isot == '2013-07-02T01:39:20.000'
        assert header.frame_nbytes == 64000 + 4096
        assert header.payload_nbytes == 64000
        assert header.mutable is False
        with pytest.raises(TypeError):
            header['NCHAN'] = 2
        assert header['NCHAN'] == 1
        with pytest.raises(AttributeError):
            header.python

        with open(str(tmpdir.join('test.dada')), 'w+b') as s:
            header.tofile(s)
            assert s.tell() == header.nbytes
            s.seek(0)
            header2 = dada.DADAHeader.fromfile(s)
            assert header2 == header
            assert header2.mutable is False
            assert s.tell() == header.nbytes

        with open(str(tmpdir.join('test.dada')), 'w+b') as s:
            # Now create header with wrong HDR_SIZE in file.
            bad_header = header.copy()
            bad_header['HDR_SIZE'] = 1000
            with pytest.raises(ValueError):
                bad_header.tofile(s)
            # Now write header explicitly, skipping the check in tofile,
            # so we create a bad header.
            for line in bad_header._tolines():
                s.write((line + '\n').encode('ascii'))
            s.write('# end of header\n'.encode('ascii'))
            s.seek(0)
            with pytest.warns(UserWarning, match='Odd'):
                dada.DADAHeader.fromfile(s)

        with open(SAMPLE_FILE, 'rb') as fh, \
                open(str(tmpdir.join('test2.dada')), 'w+b') as s:
            # And an incomplete header for completeness.
            s.write(fh.read(1000))
            s.seek(0)
            with pytest.raises(EOFError):
                dada.DADAHeader.fromfile(s)

        # Note that this is not guaranteed to preserve order!
        header3 = dada.DADAHeader.fromkeys(**header)
        assert header3 == header
        assert header3.mutable is True
        # Check attribute setting.
        header3.start_time = header.start_time - 0.5 * u.day
        assert np.abs(header3.start_time
                      - (header.start_time - 0.5 * u.day)) < 1 * u.ns
        assert np.abs(header3.time - (header.time - 0.5 * u.day)) < 1 * u.ns
        # Check against rounding.
        just_below_int = Time(55000, -1e-15, format='mjd')
        header3.start_time = just_below_int
        assert header3['MJD_START'] == '54999.999999999999999'
        header3['NCHAN'] = 2
        assert header3['NCHAN'] == 2
        header3.frame_nbytes = 9096
        assert header3.payload_nbytes == 5000
        # Try initialising with properties instead of keywords.
        # Here, we first just try the start time and offset.
        header4 = dada.DADAHeader.fromvalues(
            start_time=header.start_time,
            offset=header.time - header.start_time,
            bps=header.bps, complex_data=header.complex_data,
            sample_rate=header.sample_rate, sideband=header.sideband,
            samples_per_frame=header.samples_per_frame,
            sample_shape=header.sample_shape,
            source=header['SOURCE'], ra=header['RA'], dec=header['DEC'],
            telescope=header['TELESCOPE'], instrument=header['INSTRUMENT'],
            receiver=header['RECEIVER'], freq=header['FREQ'],
            pic_version=header['PIC_VERSION'])
        assert header4 == header
        assert header4.mutable is True
        # And now try the time and offset.
        header5 = dada.DADAHeader.fromvalues(
            offset=header.time - header.start_time, time=header.time,
            bps=header.bps, complex_data=header.complex_data,
            sample_rate=header.sample_rate, sideband=header.sideband,
            samples_per_frame=header.samples_per_frame,
            sample_shape=header.sample_shape,
            source=header['SOURCE'], ra=header['RA'], dec=header['DEC'],
            telescope=header['TELESCOPE'], instrument=header['INSTRUMENT'],
            receiver=header['RECEIVER'], freq=header['FREQ'],
            pic_version=header['PIC_VERSION'])
        # Finally try the time and start time.
        assert header5 == header
        header6 = dada.DADAHeader.fromvalues(
            time=header.time, start_time=header.start_time,
            bps=header.bps, complex_data=header.complex_data,
            sample_rate=header.sample_rate, sideband=header.sideband,
            samples_per_frame=header.samples_per_frame,
            sample_shape=header.sample_shape,
            source=header['SOURCE'], ra=header['RA'], dec=header['DEC'],
            telescope=header['TELESCOPE'], instrument=header['INSTRUMENT'],
            receiver=header['RECEIVER'], freq=header['FREQ'],
            pic_version=header['PIC_VERSION'])
        assert header6 == header
        # Check repr can be used to instantiate header
        header7 = eval('dada.' + repr(header))
        assert header7 == header
        # repr includes the comments
        assert header7.comments == header.comments
        # Therefore repr should be identical too.
        assert repr(header7) == repr(header)
        # Check instantiation via tuple
        header8 = dada.DADAHeader(((key, (header[key], header.comments[key]))
                                   for key in header))
        assert header8 == header
        assert header8.comments == header.comments
        # Check copying
        header9 = header.copy()
        assert header9 == header
        assert header9.mutable is True
        assert header9.comments == header.comments
        header10 = copy.copy(header9)
        assert header10 == header
        assert header10.mutable is True
        assert header10.comments == header.comments

    def test_header_impossible_samples_per_frame(self):
        with pytest.raises(ValueError):
            dada.DADAHeader.fromvalues(nchan=1, npol=1, complex_data=False,
                                       bps=4, samples_per_frame=10001)

    def test_pickle_header(self, tmpdir):
        with open(SAMPLE_FILE, 'rb') as fh:
            header = dada.DADAHeader.fromfile(fh)

        pickled = pickle.dumps(header)
        recovered = pickle.loads(pickled)
        assert isinstance(recovered, dada.DADAHeader)
        assert recovered == header

    def test_payload(self, tmpdir):
        payload = self.payload
        assert payload.nbytes == 64000
        assert payload.shape == (16000, 2, 1)
        # Check sample shape validity.
        assert payload.sample_shape == (2, 1)
        assert payload.sample_shape.npol == 2
        assert payload.sample_shape.nchan == 1
        assert payload.size == 32000
        assert payload.ndim == 3
        assert payload.dtype == np.complex64
        assert np.all(payload[:3] == np.array(
            [[[-38.-38.j], [-38.-38.j]],
             [[-38.-38.j], [-40.+0.j]],
             [[-105.+60.j], [85.-15.j]]], dtype=np.complex64))

        with open(str(tmpdir.join('test.dada')), 'w+b') as s:
            payload.tofile(s)
            s.seek(0)
            payload2 = dada.DADAPayload.fromfile(s, payload_nbytes=64000,
                                                 sample_shape=(2, 1), bps=8,
                                                 complex_data=True)
            assert payload2 == payload
            assert s.tell() == 64000
            with pytest.raises(EOFError):
                # Too few bytes.
                s.seek(100)
                dada.DADAPayload.fromfile(s, self.header, memmap=False)
        payload3 = dada.DADAPayload.fromdata(payload.data, bps=8)
        assert payload3 == payload
        with open(SAMPLE_FILE, 'rb') as fh:
            fh.seek(4096)
            payload4 = dada.DADAPayload.fromfile(fh, self.header, memmap=True)
            assert fh.tell() == 4096+payload4.nbytes
        assert isinstance(payload4.words, np.memmap)
        assert not isinstance(payload.words, np.memmap)
        assert payload == payload4

    def test_file_reader(self):
        with dada.open(SAMPLE_FILE, 'rb') as fh:
            header = fh.read_header()
            assert header == self.header
            current_pos = fh.tell()
            frame_rate = fh.get_frame_rate()
            assert frame_rate == header.sample_rate / header.samples_per_frame
            assert fh.tell() == current_pos
            # Frame is done below, as is writing in binary.

    def test_frame(self, tmpdir):
        with dada.open(SAMPLE_FILE, 'rb') as fh:
            frame = fh.read_frame(memmap=False)
        header, payload = frame.header, frame.payload
        assert header == self.header
        assert payload == self.payload
        assert frame == dada.DADAFrame(header, payload)
        assert frame.shape == payload.shape
        assert frame.size == payload.size
        assert frame.ndim == payload.ndim
        assert np.all(frame[:3] == np.array(
            [[[-38.-38.j], [-38.-38.j]],
             [[-38.-38.j], [-40.+0.j]],
             [[-105.+60.j], [85.-15.j]]], dtype=np.complex64))
        with open(str(tmpdir.join('test.dada')), 'w+b') as s:
            frame.tofile(s)
            s.seek(0)
            frame2 = dada.DADAFrame.fromfile(s, memmap=False)
            assert s.tell() == frame.nbytes
        assert frame2 == frame
        frame3 = dada.DADAFrame.fromdata(payload.data, header)
        assert frame3 == frame
        frame4 = dada.DADAFrame.fromdata(payload.data, **header)
        assert frame4 == frame
        header5 = header.copy()
        frame5 = dada.DADAFrame(header5, payload, valid=False)
        assert frame5.valid is False
        assert np.all(frame5.data == 0.)
        frame5.valid = True
        assert frame5 == frame

    def test_frame_memmap(self, tmpdir):
        # Get frame regular way.
        with dada.open(SAMPLE_FILE, 'rb') as fr:
            frame = fr.read_frame(memmap=False)
        assert not isinstance(frame.payload.words, np.memmap)
        # Check that if we map it instead, we get the same result.
        with dada.open(SAMPLE_FILE, 'rb') as fh:
            frame2 = fh.read_frame(memmap=True)
        assert frame2 == frame
        assert isinstance(frame2.payload.words, np.memmap)
        # Bit superfluous perhaps, but check decoding as well.
        assert np.all(frame2[:3] == np.array(
            [[[-38.-38.j], [-38.-38.j]],
             [[-38.-38.j], [-40.+0.j]],
             [[-105.+60.j], [85.-15.j]]], dtype=np.complex64))
        assert np.all(frame2.data == frame.data)

        # Now check writing.  First, without memmap, just ensuring writing
        # to file works as well as to BytesIO done above.
        filename = str(tmpdir.join('a.dada'))
        with dada.open(filename, 'wb') as fw:
            fw.write_frame(frame)

        with dada.open(filename, 'rb') as fw:
            frame3 = fw.read_frame()

        assert frame3 == frame
        # Now memmap file to be written to.
        filename2 = str(tmpdir.join('a2.dada'))
        with dada.open(filename2, 'wb') as fw:
            frame4 = fw.memmap_frame(frame.header)
        # Initially no data set, so frames should not match yet.
        assert frame4 != frame
        # So, if we read this file, it also should not match
        with dada.open(filename2, 'rb') as fw:
            frame5 = fw.read_frame()
        assert frame5 != frame

        # Fill in some data.  This should only update some words.
        frame4[:20] = frame[:20]
        assert np.all(frame4[:20] == frame[:20])
        assert frame4 != frame
        # Update the rest, so it becomes the same.
        frame4[20:] = frame[20:]
        assert frame4 == frame
        # delete to flush to disk just to be sure, then read and check it.
        del frame4
        with dada.open(filename, 'rb') as fw:
            frame6 = fw.read_frame()

        assert frame6 == frame
        # Some further tests for completeness;
        # initiate frame using data and header keywords.
        filename3 = str(tmpdir.join('a3.dada'))
        with dada.open(filename3, 'wb') as fw:
            fw.write_frame(self.payload.data, **self.header)
        with dada.open(filename3, 'rb') as fh:
            frame7 = fh.read_frame()
        assert frame7 == frame
        # memmap frame using header keywords.
        filename4 = str(tmpdir.join('a4.dada'))
        with dada.open(filename4, 'wb') as fw:
            frame8 = fw.memmap_frame(**self.header)
            frame8[:] = self.payload.data
        assert frame8 == frame
        del frame8
        with dada.open(filename4, 'rb') as fh:
            frame9 = fh.read_frame()
        assert frame9 == frame

    def test_filestreamer(self, tmpdir):
        start_time = self.header.time
        with dada.open(SAMPLE_FILE, 'rs') as fh:
            assert fh.header0 == self.header
            assert fh.sample_shape == (2,)
            assert fh.shape == (16000,) + fh.sample_shape
            assert fh.size == np.prod(fh.shape)
            assert fh.ndim == len(fh.shape)
            assert fh.start_time == start_time
            assert fh.sample_rate == 16 * u.MHz
            record1 = fh.read(12)
            assert fh.tell() == 12
            fh.seek(10000)
            record2 = np.zeros((2, 2), dtype=np.complex64)
            record2 = fh.read(out=record2)
            assert fh.tell() == 10002
            assert fh.time == fh.tell(unit='time')
            assert (np.abs(fh.time - (start_time + 10002 / (16 * u.MHz)))
                    < 1. * u.ns)
            fh.seek(fh.start_time + 1000 / (16*u.MHz))
            assert fh.tell() == 1000
            assert fh._last_header == fh.header0
            assert np.abs(fh.stop_time
                          - (start_time + 16000 / (16.*u.MHz))) < 1.*u.ns
            # Test seeker works with both int and str values for whence.
            assert fh.seek(13, 0) == fh.seek(13, 'start')
            assert fh.seek(-13, 2) == fh.seek(-13, 'end')
            fhseek_int = fh.seek(17, 1)
            fh.seek(-17, 'current')
            fhseek_str = fh.seek(17, 'current')
            assert fhseek_int == fhseek_str
            with pytest.raises(ValueError):
                fh.seek(0, 'last')
            fh.seek(1, 'end')
            with pytest.raises(EOFError):
                fh.read()

        assert record1.shape == (12, 2)
        assert np.all(record1[:3] == np.array(
            [[-38.-38.j, -38.-38.j],
             [-38.-38.j, -40.+0.j],
             [-105.+60.j, 85.-15.j]], dtype=np.complex64))
        assert record1.shape == (12, 2) and record1.dtype == np.complex64
        assert np.all(record1 == self.payload[:12].squeeze())
        assert record2.shape == (2, 2)
        assert np.all(record2 == self.payload[10000:10002].squeeze())

        filename = str(tmpdir.join('a.dada'))
        with dada.open(filename, 'ws', header0=self.header,
                       squeeze=False) as fw:
            assert fw.sample_rate == 16 * u.MHz
            fw.write(self.payload.data)
            assert fw.start_time == start_time
            assert (np.abs(fw.time - (start_time + 16000 / (16. * u.MHz)))
                    < 1. * u.ns)

        with dada.open(filename, 'rs') as fh:
            data = fh.read()
            assert fh.start_time == start_time
            assert (np.abs(fh.time - (start_time + 16000 / (16. * u.MHz)))
                    < 1. * u.ns)
            assert fh.stop_time == fh.time
            assert fh.sample_rate == 16 * u.MHz
        assert np.all(data == self.payload.data.squeeze())

        # Try single polarisation, and check initialisation by header keywords.
        # Use a new file since the mmap used in frame3 may still have a handle
        # to the previous one, which causes (sensible) failures on windows.
        # See https://bugs.python.org/issue40720
        filename2 = str(tmpdir.join('a2.dada'))
        h = self.header
        with dada.open(filename2, 'ws', time=h.time, bps=h.bps,
                       complex_data=h.complex_data, sample_rate=h.sample_rate,
                       payload_nbytes=32000, npol=1, nchan=1) as fw:
            fw.write(self.payload.data[:, 0, 0])
            assert np.abs(fw.start_time - start_time) < 1.*u.ns
            assert (np.abs(fw.time - (start_time + 16000 / (16. * u.MHz)))
                    < 1. * u.ns)

        with dada.open(filename2, 'rs') as fh:
            data_onepol = fh.read()
            assert np.abs(fh.start_time - start_time) < 1.*u.ns
            assert np.abs(fh.stop_time
                          - (start_time + 16000 / (16.*u.MHz))) < 1.*u.ns
        assert np.all(data_onepol == self.payload.data[:, 0, 0])

        # Try reading a single polarization.
        with dada.open(SAMPLE_FILE, 'rs', subset=0) as fh:
            assert fh.sample_shape == ()
            assert fh.subset == (0,)
            record3 = fh.read(12)
            assert np.all(record3 == record1[:12, 0])

        # Create an npol=2, nchan=2 file and subset it.
        data2d = np.array([data, -data]).transpose(1, 2, 0)
        filename3 = str(tmpdir.join('a3.dada'))
        with dada.open(filename3, 'ws', time=h.time, bps=h.bps,
                       complex_data=h.complex_data, sample_rate=h.sample_rate,
                       payload_nbytes=32000, npol=2, nchan=2) as fw:
            fw.write(data2d)
            assert np.abs(fw.start_time - start_time) < 1.*u.ns
            assert (np.abs(fw.time - (start_time + 16000 / (16. * u.MHz)))
                    < 1. * u.ns)

        # First check if write was successful.
        with dada.open(filename3, 'rs') as fh:
            assert fh.sample_shape == (2, 2)
            data_all = fh.read()
            assert np.all(data_all == data2d)

        # Then read right polarization, but read channels in reverse order.
        with dada.open(filename3, 'rs', subset=(1, [1, 0])) as fh:
            assert fh.sample_shape == (2,)
            data_sub = fh.read()
            assert np.all(data_sub[:, 1] == data2d[:, 1, 0])

        # Test that squeeze attribute works on read (including in-place read;
        # we implicitly tested squeeze=True above).
        with dada.open(SAMPLE_FILE, 'rs', squeeze=False) as fh:
            assert fh.sample_shape == (2, 1)
            assert fh.sample_shape.npol == 2
            assert fh.sample_shape.nchan == 1
            assert fh.read(1).shape == (1, 2, 1)
            assert fh.read(10).shape == (10, 2, 1)
            fh.seek(0)
            out = np.zeros((12, 2, 1), dtype=np.complex64)
            fh.read(out=out)
            assert fh.tell() == 12
            assert np.all(out.squeeze() == record1)

        # Test that squeeze attribute works on write.
        dada_test_squeeze = str(tmpdir.join('test_squeeze.dada'))
        with dada.open(dada_test_squeeze, 'ws', header0=self.header) as fw:
            assert fw.sample_shape == (2,)
            assert fw.sample_shape.npol == 2
            fw.write(self.payload.data.squeeze())
        dada_test_nosqueeze = str(tmpdir.join('test_nosqueeze.dada'))
        with dada.open(dada_test_nosqueeze, 'ws', header0=self.header,
                       squeeze=False) as fw:
            assert fw.sample_shape == (2, 1)
            assert fw.sample_shape.npol == 2
            assert fw.sample_shape.nchan == 1
            fw.write(self.payload.data)

        with dada.open(dada_test_squeeze, 'rs', squeeze=False) as fhs, \
                dada.open(dada_test_nosqueeze, 'rs', squeeze=False) as fhns:
            assert np.all(fhs.read() == self.payload)
            assert np.all(fhns.read() == self.payload)

        # Test passing squeeze=False along with header keywords (otherwise
        # identical to test above).
        h = self.header

        filename4 = str(tmpdir.join('a4.dada'))
        with dada.open(filename4, 'ws', time=h.time,
                       payload_nbytes=h.payload_nbytes,
                       sample_rate=h.sample_rate, complex_data=h.complex_data,
                       sample_shape=h.sample_shape, bps=h.bps,
                       squeeze=False) as fw:
            assert fw.sample_shape == (2, 1)
            assert fw.sample_shape.npol == 2
            assert fw.sample_shape.nchan == 1
            fw.write(self.payload.data)

        with dada.open(dada_test_squeeze, 'rs', squeeze=False) as fhs, \
                dada.open(dada_test_nosqueeze, 'rs', squeeze=False) as fhns:
            assert np.all(fhs.read() == self.payload)
            assert np.all(fhns.read() == self.payload)

    def test_incomplete_stream(self, tmpdir):
        """Test that writing an incomplete stream is possible, and that frame
        set is valid but invalid samples use the fill value.
        """
        filename = str(tmpdir.join('a.dada'))
        with pytest.warns(UserWarning, match='partial buffer'):
            with dada.open(filename, 'ws', header0=self.header,
                           squeeze=False) as fw:
                fw.write(self.payload[:10])

        with dada.open(filename, 'rs', squeeze=False) as fwr:
            data = fwr.read()
            assert np.all(data[:10] == self.payload[:10])
            assert np.all(data[10:] == fwr.fill_value)

    def test_pickle(self):
        # Only simple tests here; more complete ones in vdif.
        with dada.open(SAMPLE_FILE, 'rs', squeeze=False) as fh:
            fh.seek(6)
            pickled = pickle.dumps(fh)
            fh.read(3)
            with pickle.loads(pickled) as fh2:
                assert fh2.tell() == 6
                fh2.read(10)

            assert fh.tell() == 9

        with pickle.loads(pickled) as fh3:
            assert fh3.tell() == 6
            fh3.read(1)

        closed = pickle.dumps(fh)
        with pickle.loads(closed) as fh4:
            assert fh4.closed
            with pytest.raises(ValueError):
                fh4.read(1)

    def test_multiple_files_stream(self, tmpdir):
        start_time = self.header.time
        data = self.payload.data.squeeze()
        header = self.header.copy()
        header.payload_nbytes = self.header.payload_nbytes // 2
        filenames = (str(tmpdir.join('a.dada')),
                     str(tmpdir.join('b.dada')))
        with dada.open(filenames, 'ws', **header) as fw:
            start_time = fw.start_time
            fw.write(data[:1000])
            time1000 = fw.time
            fw.write(data[1000:])
            stop_time = fw.time
        assert start_time == header.time
        assert np.abs(time1000 - (start_time + 1000 / (16.*u.MHz))) < 1.*u.ns
        assert np.abs(stop_time - (start_time + 16000 / (16.*u.MHz))) < 1.*u.ns

        with dada.open(filenames[1], 'rs') as fr:
            assert (np.abs(fr.time - (start_time + 8000 / (16. * u.MHz)))
                    < 1. * u.ns)
            data1 = fr.read()
        assert np.all(data1 == data[8000:])

        with dada.open(filenames, 'rs') as fr:
            assert fr.start_time == start_time
            assert fr.time == start_time
            assert np.abs(fr.stop_time
                          - (start_time + 16000 / (16. * u.MHz))) < 1. * u.ns
            data2 = fr.read()
            assert fr.time == fr.stop_time
        assert np.all(data2 == data)

        # Pass sequentialfile objects to reader.
        filenames = (str(tmpdir.join('a2.dada')),
                     str(tmpdir.join('b2.dada')))
        with sf.open(filenames, 'w+b',
                     file_size=(header.payload_nbytes + 4096)) as fraw, \
                dada.open(fraw, 'ws', header0=header) as fw:
            fw.write(data)

        with dada.open(filenames, 'rs') as fr:
            data3 = fr.read()
        assert np.all(data3 == data)

        # Test passing stream reader kwargs.
        with dada.open(filenames, 'rs', subset=1, squeeze=False) as fr:
            data4 = fr.read()
        assert np.all(data4.squeeze() == data[:, 1])

        # Test pickling.
        with dada.open(filenames, 'rs', subset=1, squeeze=False) as fr:
            fr.seek(10)
            pickled = pickle.dumps(fr)

        with pickle.loads(pickled) as fr2:
            assert fr2.tell() == 10
            datap = fr2.read()
        assert np.all(datap.squeeze() == data[10:, 1])

        # Check that we can't pass a filename sequence in 'wb' mode.
        with pytest.raises(ValueError):
            dada.open(filenames, 'wb')

    def test_partial_last_frame(self, tmpdir):
        """Test reading an incomplete frame from one or a sequence of files."""
        # Prepare file sequence.
        header = self.header.copy()
        data = self.payload.data.squeeze()
        data = np.r_[data, data, data]
        filenames = [str(tmpdir.join('a.dada')),
                     str(tmpdir.join('b.dada')),
                     str(tmpdir.join('c.dada'))]
        with dada.open(filenames, 'ws', header0=header) as fw:
            fw.write(data)

        # Replace c.dada with partially complete file.
        with dada.open(str(tmpdir.join('c.dada')), 'rb') as fh, \
                dada.open(str(tmpdir.join('c_partial.dada')), 'wb') as fw:
            full_filesize = fh.seek(0, 2)
            fh.seek(0)
            fw.write(fh.read(full_filesize // 2 - 37))

        filenames[-1] = str(tmpdir.join('c_partial.dada'))

        # Check reading single partial frame.
        with dada.open(filenames[-1], 'rs') as fh:
            # Check that we've written the right number of bytes to file.
            filesize = fh.fh_raw.seek(0, 2)
            fh.fh_raw.seek(0)
            assert filesize == full_filesize // 2 - 37
            # Payload truncates 3 bytes so there is an integer number of
            # complete samples.
            assert fh.header0.frame_nbytes == filesize - 3
            assert fh.header0.nbytes == self.header.nbytes
            assert fh.samples_per_frame == (
                (filesize - self.header.nbytes) * 8 // fh.header0.bps // 2
                // np.prod(fh.header0.sample_shape))
            assert fh.header0 is fh._last_header
            assert np.abs(fh.stop_time - fh.start_time
                          - 7478 / fh.sample_rate) < 1 * u.ns
            assert fh.shape == (7478, 2)
            # Taking advantage of data being repeated 3 times.
            assert np.all(fh.read() == data[:7478])

        # Check reading sequence of files.
        with dada.open(filenames) as fh:
            assert fh.samples_per_frame == header.samples_per_frame
            assert np.abs(fh.stop_time
                          - fh.start_time - 39478 / fh.sample_rate) < 1 * u.ns
            assert fh.shape == (39478, 2)
            assert np.all(fh.read() == data[:39478])
            fh.seek(-29, 2)
            assert np.all(fh.read() == data[7478 - 29:7478])
            assert fh.tell() == 39478

        # Replace c.dada with only the header (and no payload).
        with dada.open(str(tmpdir.join('c.dada')), 'rb') as fh, \
                dada.open(str(tmpdir.join('c_header_only.dada')), 'wb') as fw:
            fw.write(fh.read(4096))
            fh.seek(0)
            header_c = fh.read_header()

        filenames[-1] = str(tmpdir.join('c_header_only.dada'))

        # Check that reading the frame gives payload of zero size.
        with dada.open(str(tmpdir.join('c_header_only.dada')), 'rb') as fp:
            assert fp.read_header() == header_c
            fp.seek(0)
            with pytest.raises(Exception) as excinfo:
                fp.read_frame()
            # Different exceptions for linux and windows.
            assert ("mmap length is greater" in str(excinfo.value)
                    or "Not enough memory resources" in str(excinfo.value))

        with pytest.raises(EOFError) as excinfo:
            with dada.open(str(tmpdir.join('c_header_only.dada')), 'rs') as fp:
                pass
        assert "appears to end without" in str(excinfo.value)

        # Reading the new sequence, the last frame should be ignored.
        with dada.open(filenames) as fh:
            assert np.abs(fh.stop_time
                          - fh.start_time - 32000 / fh.sample_rate) < 1 * u.ns
            assert fh.shape == (32000, 2)

    def test_template_stream(self, tmpdir):
        start_time = self.header.time
        data = self.payload.data.squeeze()
        header = self.header.copy()
        header.payload_nbytes = self.header.payload_nbytes // 4
        template = str(tmpdir.join('a{frame_nr}.dada'))
        with dada.open(template, 'ws', header0=header) as fw:
            fw.write(data[:1000])
            time1000 = fw.time
            fw.write(data[1000:])
            stop_time = fw.time
        assert np.abs(time1000 - (header.time + 1000 / (16.*u.MHz))) < 1.*u.ns
        assert np.abs(stop_time
                      - (header.time + 16000 / (16. * u.MHz))) < 1. * u.ns

        with dada.open(template.format(frame_nr=1), 'rs') as fr:
            data1 = fr.read()
            assert fr.time == fr.stop_time
            assert np.abs(fr.start_time
                          - (start_time + 4000 / (16.*u.MHz))) < 1.*u.ns
            assert np.abs(fr.stop_time
                          - (start_time + 8000 / (16.*u.MHz))) < 1.*u.ns
        assert np.all(data1 == data[4000:8000])

        with dada.open(template, 'rs') as fr:
            assert fr.time == start_time
            data2 = fr.read()
            assert fr.stop_time == fr.time
            assert np.abs(fr.stop_time
                          - (header.time + 16000 / (16. * u.MHz))) < 1. * u.ns
        assert np.all(data2 == data)

    @pytest.mark.skipif(os.name == 'nt',
                        reason='Windows does not allow colon in file names')
    def test_complicated_template_stream(self, tmpdir):
        # More complicated template, 8 files.
        start_time = self.header.time
        data = self.payload.data.squeeze()
        header = self.header.copy()
        header.payload_nbytes = self.header.payload_nbytes // 8
        template = str(tmpdir
                       .join('{utc_start}_{obs_offset:016d}.000000.dada'))
        with dada.open(template, 'ws', header0=header) as fw:
            fw.write(data[:7000])
            assert fw.start_time == header.time
            assert (np.abs(fw.time - (start_time + 7000 / (16. * u.MHz)))
                    < 1. * u.ns)
            fw.write(data[7000:])
            assert (np.abs(fw.time - (start_time + 16000 / (16. * u.MHz)))
                    < 1. * u.ns)

        name3 = template.format(utc_start=header['UTC_START'],
                                obs_offset=header['OBS_OFFSET']
                                + 3 * header.payload_nbytes)
        with dada.open(name3, 'rs') as fr:
            assert np.abs(fr.start_time
                          - (start_time + 6000 / (16.*u.MHz))) < 1.*u.ns
            assert np.abs(fr.stop_time
                          - (start_time + 8000 / (16.*u.MHz))) < 1.*u.ns
            data1 = fr.read()
            assert fr.stop_time == fr.time
        assert np.all(data1 == data[6000:8000])

        # We cannot just open using the same template, since UTC_START is
        # not available.
        with pytest.raises(KeyError):
            dada.open(template, 'rs')

        kwargs = dict(UTC_START=header['UTC_START'],
                      OBS_OFFSET=header['OBS_OFFSET']
                      + 3 * header.payload_nbytes,
                      FILE_SIZE=header['FILE_SIZE'])
        with dada.open(template, 'rs', **kwargs) as fr:
            assert (np.abs(fr.time - (start_time + 6000 / (16. * u.MHz)))
                    < 1. * u.ns)
            data2 = fr.read()
            assert fr.time == fr.stop_time
            assert np.abs(fr.stop_time
                          - (start_time + 16000 / (16.*u.MHz))) < 1.*u.ns
        assert np.all(data2 == data[6000:])

        # Just to check internal checks are OK.
        with pytest.raises(ValueError):
            dada.open(name3, 's')
        with pytest.raises(TypeError):
            # Extraneous argument.
            dada.open(name3, 'rs', files=(name3,))


class TestDADAFileNameSequencer:
    def setup(self):
        with open(SAMPLE_FILE, 'rb') as fh:
            self.header = dada.DADAHeader.fromfile(fh)

    def test_offset_enumeration(self):
        fns = DADAFileNameSequencer(
            '{obs_offset:06d}.x', {'OBS_OFFSET': 10, 'FILE_SIZE': 20})
        assert fns[0] == '000010.x'
        assert fns[9] == '000190.x'

        with pytest.raises(KeyError):
            DADAFileNameSequencer('{obs_offset:06d}.x', {'OBS_OFFSET': 10})

    def test_complicated_enumeration(self):
        # Tests that frame_nr properly draws from file_nr.
        template = '{frame_nr}_{obs_offset:016d}.dada'
        fns = DADAFileNameSequencer(template, self.header)
        assert fns[0] == '0_0000006400000000.dada'
        assert fns[1] == '1_0000006400064000.dada'
        assert fns[10] == '10_0000006400640000.dada'

        # Follow the typical naming scheme:
        # 2016-04-23-07:29:30_0000000000000000.000000.dada
        template = '{utc_start}_{obs_offset:016d}.000000.dada'
        fns = DADAFileNameSequencer(template, self.header)
        assert fns[0] == '2013-07-02-01:37:40_0000006400000000.000000.dada'
        assert fns[100] == '2013-07-02-01:37:40_0000006406400000.000000.dada'



# Licensed under the GPLv3 - see LICENSE
import bisect
import operator


class RawOffsets:
    """File offset tracker.

    Keeps track of offsets from expected file position as a
    function of frame number, by keeping a joint list, of the
    first frame number beyond which a certain offset will hold.

    The offset for a given frame number is retrieved via ``__getitem__``,
    while new offsets are added via ``__setitem__`` (keeping the
    list of frame numbers minimal for identical offsets).

    Parameters
    ----------
    frame_nr : list
        Frame number for which one has offsets.
    offset : list
        Corresponding offsets.
    frame_nbytes : int
        Possible frame size to include in all returned offsets, i.e.,
        output will be ``offset + index * frame_nbytes``.  Default: 0.

    Examples
    --------
    The usage is best seen through an example::

      >>> from baseband.base.offsets import RawOffsets
      >>> offsets = RawOffsets([6], [5])
      >>> offsets[3]  # Implicitly 0 before first entry
      0
      >>> offsets[10]  # Assumed the same as 6
      5
      >>> offsets[10] = 9  # But suppose we find 10 has 9.
      >>> offsets[10]  # Then it takes that
      9
      >>> offsets[9]  # But before is still 5.
      5
      >>> offsets[8] = 9  # But if we find 8 has 9 too.
      >>> offsets[9]  # Then 9 is the same.
      9
      >>> offsets  # And the list is kept minimal.
      RawOffsets(frame_nr=[6, 8], offset=[5, 9], frame_nbytes=0)
      >>> offsets[9] = 9  # This is a no-op.
      >>> offsets[10] = 10  # But this is not.
      >>> offsets
      RawOffsets(frame_nr=[6, 8, 10], offset=[5, 9, 10], frame_nbytes=0)

    Similarly, if a frame size is passed in::

      >>> offsets = RawOffsets([6, 8, 10], [5, 9, 10], frame_nbytes=1000)
      >>> offsets
      RawOffsets(frame_nr=[6, 8, 10], offset=[5, 9, 10], frame_nbytes=1000)
      >>> offsets[1]
      1000
      >>> offsets[8]
      8009
      >>> offsets[8] = 8005  # This removes the offset for 9.
      >>> offsets[8]
      8005
      >>> offsets
      RawOffsets(frame_nr=[6, 10], offset=[5, 10], frame_nbytes=1000)

    """

    def __init__(self, frame_nr=None, offset=None, frame_nbytes=0):
        if frame_nr is None:
            frame_nr = []
        if offset is None:
            offset = []
        if len(frame_nr) != len(offset):
            raise ValueError('must have equal number of frame numbers '
                             'and offsets.')
        self.frame_nr = frame_nr
        self.offset = offset
        self.frame_nbytes = operator.index(frame_nbytes)

    def __getitem__(self, frame_nr):
        # Keep default case of no offsets as fast as possible.
        base = frame_nr * self.frame_nbytes
        if not self.frame_nr:
            return base
        # Find first index for which frame_nr < value-at-index,
        # hence the offset at the previous index is the one we need.
        index = bisect.bisect_right(self.frame_nr, frame_nr)
        return base if index == 0 else base + self.offset[index - 1]

    def __setitem__(self, frame_nr, offset):
        # Get the offset from expected.
        offset -= frame_nr * self.frame_nbytes
        # Find first index for which frame_nr < value-at-index.
        # Hence, this is where we should be if we do not yet exist.
        index = bisect.bisect_right(self.frame_nr, frame_nr)
        # If the entry already exists (should not really happen),
        # and the new value is different, replace it.
        if index > 0 and self.frame_nr[index-1] == frame_nr:
            if self.offset[index-1] == offset:
                return

            # Best to *remove* the entry, since the new value may
            # be consistent with one of the surrounding values,
            # in which case we can shorten our list.
            self.frame_nr.pop(index-1)
            self.offset.pop(index-1)
            index -= 1

        # If the offset at the next location is the same as ours,
        # then we can keep everything consistent by just moving the
        # frame number to our value.
        if index < len(self) and self.offset[index] == offset:
            self.frame_nr[index] = frame_nr
        elif index == 0 or self.offset[index-1] != offset:
            # Otherwise, if we add new information, insert ourserlves.
            self.frame_nr.insert(index, frame_nr)
            self.offset.insert(index, offset)

    def __len__(self):
        return len(self.frame_nr)

    def __repr__(self):
        return ('{0}(frame_nr={1}, offset={2}, frame_nbytes={3})'
                .format(type(self).__name__, self.frame_nr, self.offset,
                        self.frame_nbytes))



# Licensed under the GPLv3 - see LICENSE
from ..base.frame import FrameBase
from .header import GSBHeader
from .payload import GSBPayload


__all__ = ['GSBFrame']


class GSBFrame(FrameBase):
    """Frame encapsulating GSB rawdump or phased data.

    For rawdump data, lines in the timestamp file are associated with single
    blocks of raw data.  For phased data, the lines are associated with one or
    two polarisations, each consisting of two blocks of raw data. Hence, the
    raw data come from two or four files.

    Parameters
    ----------
    header : `~baseband.gsb.GSBHeader`
        Based on line from rawdump or phased timestamp file.
    payload : `~baseband.gsb.GSBPayload`
        Based on a single block of rawdump data, or the combined blocks for
        phased data.
    valid : bool, optional
        Whether the data are valid.  Default: `True`.
    verify : bool, optional
        Whether to verify consistency of the frame parts.  Default: `True`.

    Notes
    -----
    GSB files do not support storing whether data are valid or not on disk.
    Hence, this has to be determined independently.  If ``valid=False``, any
    decoded data are set to ``cls.fill_value`` (by default, 0).

    The Frame can also be read instantiated using class methods:

      fromfile : read header and payload from their respective filehandles

      fromdata : encode data as payload

    Of course, one can also do the opposite:

      tofile : method to write header and payload to filehandles (splitting
               payload in the appropriate files).

      data : property that yields full decoded payload

    A number of properties are defined: `shape`, `dtype` and `size` are
    the shape, type and number of complete samples of the data array, and
    `nbytes` the frame size in bytes.  Furthermore, the frame acts as a
    dictionary, with keys those of the header.  Any attribute that is not
    defined on the frame itself, such as ``.time`` will be looked up on the
    header as well.
    """
    _header_class = GSBHeader
    _payload_class = GSBPayload

    @classmethod
    def fromfile(cls, fh_ts, fh_raw, payload_nbytes=1 << 22,
                 sample_shape=(1,), bps=4, complex_data=False,
                 valid=True, verify=True):
        """Read a frame from timestamp and raw data filehandles.

        Any arguments beyond the filehandle are used to help initialize the
        payload, except for ``valid`` and ``verify``, which are passed on
        to the header and class initializers.

        Parameters
        ----------
        fh_ts : filehandle
            To the timestamp file.  The next line will be read.
        fh_raw : file_handle or tuple
            Should be a single handle for a rawdump data frame, or a tuple
            containing tuples with pairs of handles for a phased one.  E.g.,
            ``((L1, L2), (R1, R2))`` for left and right polarisations.
        payload_nbytes : int, optional
            Size of the individual payloads in bytes.  Default: ``2**22``
            (4 MB).
        sample_shape : tuple, optional
            Shape of the samples (e.g., (nchan,)).  Default: (1,).
        bps : int, optional
            Bits per elementary sample.  Default: 4.
        complex_data : bool, optional
            Whether data are complex.  Default: `False`.
        valid : bool, optional
            Whether the data are valid (default: `True`). Note that this cannot
            be inferred from the header or payload itself.  If `False`, any
            data read will be set to ``cls.fill_value``.
        verify : bool, optional
            Whether to verify consistency of the frame parts.  Default: `True`.
        """
        header = cls._header_class.fromfile(fh_ts, verify=verify)
        payload = cls._payload_class.fromfile(
            fh_raw, payload_nbytes=payload_nbytes,
            sample_shape=sample_shape, bps=bps, complex_data=complex_data)
        return cls(header, payload, valid=valid, verify=verify)

    def tofile(self, fh_ts, fh_raw):
        """Write encoded frame to timestamp and raw data filehandles.

        Parameters
        ----------
        fh_ts : filehandle
            To the timestamp file.  A line will be added to it.
        fh_raw : file_handle or tuple
            Should be a single handle for a rawdump data frame, or a tuple
            containing tuples with pairs of handles for a phased one.  E.g.,
            ``((L1, L2), (R1, R2))`` for left and right polarisations.
        """
        self.header.tofile(fh_ts)
        self.payload.tofile(fh_raw)

    @classmethod
    def fromdata(cls, data, header=None, *,
                 bps=4, valid=True, verify=True, **kwargs):
        """Construct frame from data and header.

        Parameters
        ----------
        data : `~numpy.ndarray`
            Array holding data to be encoded.
        header : ``cls._header_class``
            Header for the frame.
        bps : int, optional
            Bits per elementary sample.  Default: 4.
        valid : bool, optional
            Whether this payload contains valid data.
        verify : bool, optional
            Whether to verify the header and frame correctness.
        **kwargs :
            Used to intialize the header, if not given.
        """
        if header is None:
            header = cls._header_class.fromvalues(verify=verify, **kwargs)
        payload = cls._payload_class.fromdata(data, bps=bps)
        return cls(header, payload, valid=valid, verify=verify)

    @property
    def nbytes(self):
        """Size of the encoded frame in the raw data file in bytes."""
        return self.payload.nbytes
